"""
Code from the 3D UNet implementation:
https://github.com/wolny/pytorch-3dunet/
"""
import importlib
from functools import partial

import torch
import torch.nn as nn
from torch.nn import functional as F


def number_of_features_per_level(init_channel_number, num_levels):
    return [init_channel_number * 2**k for k in range(num_levels)]


def conv3d(in_channels, out_channels, kernel_size, bias, padding=1):
    return nn.Conv3d(in_channels, out_channels, kernel_size, padding=padding, bias=bias)


def create_conv(in_channels, out_channels, kernel_size, order, num_groups, padding=1):
    """
    Create a list of modules with together constitute a single conv layer with non-linearity
    and optional batchnorm/groupnorm.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        order (string): order of things, e.g.
            'cr' -> conv + ReLU
            'gcr' -> groupnorm + conv + ReLU
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
            'bcr' -> batchnorm + conv + ReLU
        num_groups (int): number of groups for the GroupNorm
        padding (int): add zero-padding to the input
    Return:
        list of tuple (name, module)
    """
    assert "c" in order, "Conv layer MUST be present"
    assert (
        order[0] not in "rle"
    ), "Non-linearity cannot be the first operation in the layer"
    modules = []
    for i, char in enumerate(order):
        if char == "r":
            modules.append(("ReLU", nn.ReLU(inplace=True)))
        elif char == "l":
            modules.append(
                ("LeakyReLU", nn.LeakyReLU(negative_slope=0.1, inplace=True))
            )
        elif char == "e":
            modules.append(("ELU", nn.ELU(inplace=True)))
        elif char == "c":
            # add learnable bias only in the absence of batchnorm/groupnorm
            bias = not ("g" in order or "b" in order)
            modules.append(
                ("conv", conv3d(in_channels, out_channels, kernel_size, bias, padding))
            )
        elif char == "g":
            is_before_conv = i < order.index("c")
            if is_before_conv:
                num_channels = in_channels
            else:
                num_channels = out_channels
            # use only one group if the given number of groups is greater than the number of
            # channels
            if num_channels < num_groups:
                num_groups = 1
            if num_channels % num_groups != 0:
                raise ValueError(
                    "Expected number of channels in input to be divisible by num_groups."
                    f"num_channels={num_channels}, num_groups={num_groups}"
                )
            group_norm_params = dict(num_groups=num_groups, num_channels=num_channels)
            modules.append(("groupnorm", nn.GroupNorm(**group_norm_params)))
        elif char == "b":
            is_before_conv = i < order.index("c")
            if is_before_conv:
                modules.append(("batchnorm", nn.BatchNorm3d(in_channels)))
            else:
                modules.append(("batchnorm", nn.BatchNorm3d(out_channels)))
        else:
            raise ValueError(
                f"Unsupported layer type '{char}'. MUST be one of ['b', 'g', 'r', 'l', 'e', 'c']"
            )

    return modules


class SingleConv(nn.Sequential):
    """
    Basic convolutional module consisting of a Conv3d, non-linearity and optional batchnorm /
    groupnorm. The order of operations can be specified via the `order` parameter
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        order="crg",
        num_groups=8,
        padding=1,
    ):
        super(SingleConv, self).__init__()
        for name, module in create_conv(
            in_channels, out_channels, kernel_size, order, num_groups, padding=padding
        ):
            self.add_module(name, module)


class DoubleConv(nn.Sequential):
    """
    A module consisting of two consecutive convolution layers (e.g. BatchNorm3d+ReLU+Conv3d).
    We use (Conv3d+ReLU+GroupNorm3d) by default.
    This can be changed however by providing the 'order' argument, e.g. in order
    to change to Conv3d+BatchNorm3d+ELU use order='cbe'.
    Use padded convolutions to make sure that the output (H_out, W_out) is the same
    as (H_in, W_in), so that you don't have to crop in the decoder path.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        encoder (bool): if True we're in the encoder path, otherwise we're in the decoder
        kernel_size (int): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        encoder,
        kernel_size=3,
        order="crg",
        num_groups=8,
    ):
        super(DoubleConv, self).__init__()
        if encoder:
            # we're in the encoder path
            conv1_in_channels = in_channels
            conv1_out_channels = out_channels // 2
            if conv1_out_channels < in_channels:
                conv1_out_channels = in_channels
            conv2_in_channels, conv2_out_channels = conv1_out_channels, out_channels
        else:
            # we're in the decoder path, decrease the number of channels in the 1st convolution
            conv1_in_channels, conv1_out_channels = in_channels, out_channels
            conv2_in_channels, conv2_out_channels = out_channels, out_channels

        # conv1
        self.add_module(
            "SingleConv1",
            SingleConv(
                conv1_in_channels, conv1_out_channels, kernel_size, order, num_groups
            ),
        )
        # conv2
        self.add_module(
            "SingleConv2",
            SingleConv(
                conv2_in_channels, conv2_out_channels, kernel_size, order, num_groups
            ),
        )


class ExtResNetBlock(nn.Module):
    """
    Basic UNet block consisting of a SingleConv followed by the residual block.
    The SingleConv takes care of increasing/decreasing the number of channels and also ensures that
    the number of output channels is compatible with the residual block that follows.
    This block can be used instead of standard DoubleConv in the Encoder module.
    Motivated by: https://arxiv.org/pdf/1706.00120.pdf
    Notice we use ELU instead of ReLU (order='cge') and put non-linearity after the groupnorm.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        order="cge",
        num_groups=8,
        **kwargs,
    ):
        super(ExtResNetBlock, self).__init__()
        # first convolution
        self.conv1 = SingleConv(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            order=order,
            num_groups=num_groups,
        )
        # residual block
        self.conv2 = SingleConv(
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            order=order,
            num_groups=num_groups,
        )
        # remove non-linearity from the 3rd convolution since it's going to be applied after adding
        # the residual
        n_order = order
        for c in "rel":
            n_order = n_order.replace(c, "")
        self.conv3 = SingleConv(
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            order=n_order,
            num_groups=num_groups,
        )

        # create non-linearity separately
        if "l" in order:
            self.non_linearity = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        elif "e" in order:
            self.non_linearity = nn.ELU(inplace=True)
        else:
            self.non_linearity = nn.ReLU(inplace=True)

    def forward(self, x):
        # apply first convolution and save the output as a residual
        out = self.conv1(x)
        residual = out

        # residual block
        out = self.conv2(out)
        out = self.conv3(out)

        out += residual
        out = self.non_linearity(out)

        return out


class Encoder(nn.Module):
    """
    A single module from the encoder path consisting of the optional max
    pooling layer (one may specify the MaxPool kernel_size to be different
    than the standard (2,2,2), e.g. if the volumetric data is anisotropic
    (make sure to use complementary scale_factor in the decoder path) followed by
    a DoubleConv module.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        conv_kernel_size (int): size of the convolving kernel
        apply_pooling (bool): if True use MaxPool3d before DoubleConv
        pool_kernel_size (tuple): the size of the window to take a max over
        pool_type (str): pooling layer: 'max' or 'avg'
        basic_module(nn.Module): either ResNetBlock or DoubleConv
        conv_layer_order (string): determines the order of layers
            in `DoubleConv` module. See `DoubleConv` for more info.
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        conv_kernel_size=3,
        apply_pooling=True,
        pool_kernel_size=(2, 2, 2),
        pool_type="max",
        basic_module=DoubleConv,
        conv_layer_order="crg",
        num_groups=8,
    ):
        super(Encoder, self).__init__()
        assert pool_type in ["max", "avg"]
        if apply_pooling:
            if pool_type == "max":
                self.pooling = nn.MaxPool3d(kernel_size=pool_kernel_size)
            else:
                self.pooling = nn.AvgPool3d(kernel_size=pool_kernel_size)
        else:
            self.pooling = None

        self.basic_module = basic_module(
            in_channels,
            out_channels,
            encoder=True,
            kernel_size=conv_kernel_size,
            order=conv_layer_order,
            num_groups=num_groups,
        )

    def forward(self, x):
        if self.pooling is not None:
            x = self.pooling(x)
        x = self.basic_module(x)
        return x


class Decoder(nn.Module):
    """
    A single module for decoder path consisting of the upsampling layer
    (either learned ConvTranspose3d or nearest neighbor interpolation) followed by a basic module
    (DoubleConv or ExtResNetBlock).
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): size of the convolving kernel
        scale_factor (tuple): used as the multiplier for the image H/W/D in
            case of nn.Upsample or as stride in case of ConvTranspose3d, must reverse the MaxPool3d
            operation from the corresponding encoder
        basic_module(nn.Module): either ResNetBlock or DoubleConv
        conv_layer_order (string): determines the order of layers in `DoubleConv` module.
            See `DoubleConv` for more info.
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        scale_factor=(2, 2, 2),
        basic_module=DoubleConv,
        conv_layer_order="crg",
        num_groups=8,
        mode="nearest",
    ):
        super(Decoder, self).__init__()
        if basic_module == DoubleConv:
            # if DoubleConv is the basic_module use interpolation for upsampling and concatenation
            # joining
            self.upsampling = Upsampling(
                transposed_conv=False,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                scale_factor=scale_factor,
                mode=mode,
            )
            # concat joining
            self.joining = partial(self._joining, concat=True)
        else:
            # if basic_module=ExtResNetBlock use transposed convolution upsampling and summation
            # joining
            self.upsampling = Upsampling(
                transposed_conv=True,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                scale_factor=scale_factor,
                mode=mode,
            )
            # sum joining
            self.joining = partial(self._joining, concat=False)
            # adapt the number of in_channels for the ExtResNetBlock
            in_channels = out_channels

        self.basic_module = basic_module(
            in_channels,
            out_channels,
            encoder=False,
            kernel_size=kernel_size,
            order=conv_layer_order,
            num_groups=num_groups,
        )

    def forward(self, encoder_features, x):
        x = self.upsampling(encoder_features=encoder_features, x=x)
        x = self.joining(encoder_features, x)
        x = self.basic_module(x)
        return x

    @staticmethod
    def _joining(encoder_features, x, concat):
        if concat:
            return torch.cat((encoder_features, x), dim=1)
        else:
            return encoder_features + x


class Upsampling(nn.Module):
    """
    Upsamples a given multi-channel 3D data using either interpolation or learned transposed
    convolution.
    Args:
        transposed_conv (bool): if True uses ConvTranspose3d for upsampling, otherwise uses
            interpolation
        concat_joining (bool): if True uses concatenation joining between encoder and decoder
            features, otherwise uses summation joining (see Residual U-Net)
        in_channels (int): number of input channels for transposed conv
        out_channels (int): number of output channels for transpose conv
        kernel_size (int or tuple): size of the convolving kernel
        scale_factor (int or tuple): stride of the convolution
        mode (str): algorithm used for upsampling:
            'nearest' | 'linear' | 'bilinear' | 'trilinear' | 'area'. Default: 'nearest'
    """

    def __init__(
        self,
        transposed_conv,
        in_channels=None,
        out_channels=None,
        kernel_size=3,
        scale_factor=(2, 2, 2),
        mode="nearest",
    ):
        super(Upsampling, self).__init__()

        if transposed_conv:
            # make sure that the output size reverses the MaxPool3d from the corresponding encoder
            # D_out =
            # (D_in − 1) ×  stride[0] − 2 ×  padding[0] +  kernel_size[0] + output_padding[0]
            self.upsample = nn.ConvTranspose3d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=scale_factor,
                padding=1,
            )
        else:
            self.upsample = partial(self._interpolate, mode=mode)

    def forward(self, encoder_features, x):
        output_size = encoder_features.size()[2:]
        return self.upsample(x, output_size)

    @staticmethod
    def _interpolate(x, size, mode):
        return F.interpolate(x, size=size, mode=mode)


class FinalConv(nn.Sequential):
    """
    A module consisting of a convolution layer (e.g. Conv3d+ReLU+GroupNorm3d) and the final 1x1
    convolution which reduces the number of channels to 'out_channels'.
    with the number of output channels 'out_channels // 2' and 'out_channels' respectively.
    We use (Conv3d+ReLU+GroupNorm3d) by default.
    This can be change however by providing the 'order' argument, e.g. in order
    to change to Conv3d+BatchNorm3d+ReLU use order='cbr'.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(
        self, in_channels, out_channels, kernel_size=3, order="crg", num_groups=8
    ):
        super(FinalConv, self).__init__()

        # conv1
        self.add_module(
            "SingleConv",
            SingleConv(in_channels, in_channels, kernel_size, order, num_groups),
        )

        # in the last layer a 1×1 convolution reduces the number of output channels to out_channels
        final_conv = nn.Conv3d(in_channels, out_channels, 1)
        self.add_module("final_conv", final_conv)


class Abstract3DUNet(nn.Module):
    """
    Base class for standard and residual UNet.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output segmentation masks;
            Note that that the of out_channels might correspond to either
            different semantic classes or to different binary segmentation mask.
            It's up to the user of the class to interpret the out_channels and
            use the proper loss criterion during training (i.e. CrossEntropyLoss (multi-class)
            or BCEWithLogitsLoss (two-class) respectively)
        f_maps (int, tuple): number of feature maps at each level of the encoder; if it's an
            integer the number of feature maps is given by the geometric progression:
            f_maps * 2 ^ k, k=1,2,3,4
        final_sigmoid (bool): if True apply element-wise nn.Sigmoid after the
            final 1x1 convolution, otherwise apply nn.Softmax. MUST be True if nn.BCELoss
            (two-class) is used to train the model. MUST be False if nn.CrossEntropyLoss
            (multi-class) is used to train the model.
        basic_module: basic model for the encoder/decoder (DoubleConv, ExtResNetBlock, ....)
        layer_order (string): determines the order of layers
            in `SingleConv` module. e.g. 'crg' stands for Conv3d+ReLU+GroupNorm3d.
            See `SingleConv` for more info
        f_maps (int, tuple): if int: number of feature maps in the first conv layer of the encoder
            (default: 64); if tuple: number of feature maps at each level
        num_groups (int): number of groups for the GroupNorm
        num_levels (int): number of levels in the encoder/decoder path (applied only if f_maps is
        an int)
        is_segmentation (bool): if True (semantic segmentation problem) Sigmoid/Softmax
        normalization is applied after the final convolution; if False (regression problem)
        the normalization layer is skipped at the end
        testing (bool): if True (testing mode) the `final_activation` (if present, i.e.
            `is_segmentation=true`) will be applied as the last operation during the forward pass;
            if False the model is in training mode and the `final_activation` (even if present)
            won't be applied; default: False
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid,
        basic_module,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=False,
        testing=False,
        **kwargs,
    ):
        super(Abstract3DUNet, self).__init__()

        self.testing = testing

        if isinstance(f_maps, int):
            f_maps = number_of_features_per_level(f_maps, num_levels=num_levels)

        # create encoder path consisting of Encoder modules. Depth of the encoder is equal to
        # `len(f_maps)`
        encoders = []
        for i, out_feature_num in enumerate(f_maps):
            if i == 0:
                encoder = Encoder(
                    in_channels,
                    out_feature_num,
                    apply_pooling=False,
                    basic_module=basic_module,
                    conv_layer_order=layer_order,
                    num_groups=num_groups,
                )
            else:
                # TODO: adapt for anisotropy in the data, i.e. use proper pooling kernel to make
                # the data isotropic after 1-2 pooling operations currently pools with a constant
                # kernel: (2, 2, 2)
                encoder = Encoder(
                    f_maps[i - 1],
                    out_feature_num,
                    basic_module=basic_module,
                    conv_layer_order=layer_order,
                    num_groups=num_groups,
                )
            encoders.append(encoder)

        self.encoders = nn.ModuleList(encoders)

        # create decoder path consisting of the Decoder modules. The length of the decoder is equal
        # to `len(f_maps) - 1`
        decoders = []
        reversed_f_maps = list(reversed(f_maps))
        for i in range(len(reversed_f_maps) - 1):
            if basic_module == DoubleConv:
                in_feature_num = reversed_f_maps[i] + reversed_f_maps[i + 1]
            else:
                in_feature_num = reversed_f_maps[i]

            out_feature_num = reversed_f_maps[i + 1]
            # TODO: if non-standard pooling was used, make sure to use correct striding for
            # transpose conv currently strides with a constant stride: (2, 2, 2)
            decoder = Decoder(
                in_feature_num,
                out_feature_num,
                basic_module=basic_module,
                conv_layer_order=layer_order,
                num_groups=num_groups,
            )
            decoders.append(decoder)

        self.decoders = nn.ModuleList(decoders)

        # in the last layer a 1×1 convolution reduces the number of output
        # channels to the number of labels
        self.final_conv = nn.Conv3d(f_maps[0], out_channels, 1)

        if is_segmentation:
            # semantic segmentation problem
            if final_sigmoid:
                self.final_activation = nn.Sigmoid()
            else:
                self.final_activation = nn.Softmax(dim=1)
        else:
            # regression problem
            self.final_activation = None

    def forward(self, x):
        # encoder part
        encoders_features = []
        for encoder in self.encoders:
            x = encoder(x)
            # reverse the encoder outputs to be aligned with the decoder
            encoders_features.insert(0, x)

        # remove the last encoder's output from the list
        # !!remember: it's the 1st in the list
        encoders_features = encoders_features[1:]

        # decoder part
        for decoder, encoder_features in zip(self.decoders, encoders_features):
            # pass the output from the corresponding encoder and the output
            # of the previous decoder
            x = decoder(encoder_features, x)

        x = self.final_conv(x)

        # apply final_activation (i.e. Sigmoid or Softmax) only during prediction. During training
        # the network outputs  logits and it's up to the user to normalize it before visualising
        # with tensorboard or computing validation metric
        if self.testing and self.final_activation is not None:
            x = self.final_activation(x)

        return x


class UNet3D(Abstract3DUNet):
    """
    3DUnet model from
    `"3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation"
        <https://arxiv.org/pdf/1606.06650.pdf>`.
    Uses `DoubleConv` as a basic_module and nearest neighbor upsampling in the decoder
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        **kwargs,
    ):
        super(UNet3D, self).__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=DoubleConv,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            **kwargs,
        )


class ResidualUNet3D(Abstract3DUNet):
    """
    Residual 3DUnet model implementation based on https://arxiv.org/pdf/1706.00120.pdf.
    Uses ExtResNetBlock as a basic building block, summation joining instead
    of concatenation joining and transposed convolutions for upsampling (watch out for block
    artifacts). Since the model effectively becomes a residual net, in theory it allows for deeper
    UNet.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=5,
        is_segmentation=True,
        **kwargs,
    ):
        super(ResidualUNet3D, self).__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=ExtResNetBlock,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            **kwargs,
        )


def get_model(config):
    def _model_class(class_name):
        m = importlib.import_module("pytorch3dunet.unet3d.model")
        clazz = getattr(m, class_name)
        return clazz

    assert "model" in config, "Could not find model configuration"
    model_config = config["model"]
    model_class = _model_class(model_config["name"])
    return model_class(**model_config)


if __name__ == "__main__":
    """
    testing
    """
    in_channels = 1
    out_channels = 1
    f_maps = 32
    num_levels = 3
    model = UNet3D(
        in_channels,
        out_channels,
        f_maps=f_maps,
        num_levels=num_levels,
        layer_order="cr",
    )
    print(model)

    reso = 42

    import numpy as np

    x = np.zeros((1, 1, reso, reso, reso))
    x[:, :, int(reso / 2 - 1), int(reso / 2 - 1), int(reso / 2 - 1)] = np.nan
    x = torch.FloatTensor(x)

    out = model(x)
    print(
        "%f"
        % (torch.sum(torch.isnan(out)).detach().cpu().numpy() / (reso * reso * reso))
    )



import numpy as np
import SimpleITK as sitk
from numpy import pi
from scipy.ndimage import fourier_shift, label
from skimage.filters import threshold_otsu
from skimage.measure import regionprops


def registration_exhaustive_search(
    fixed_image,
    moving_image,
    sample_per_axis=40,
    gradient_descent=False,
    threads=1,
):
    nb_dim = fixed_image.ndim
    fixed_image = sitk.GetImageFromArray(fixed_image)
    moving_image = sitk.GetImageFromArray(moving_image)
    trans = sitk.Euler3DTransform() if nb_dim == 3 else sitk.Euler2DTransform()

    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image,
        moving_image,
        trans,
        sitk.CenteredTransformInitializerFilter.GEOMETRY,
    )

    R = sitk.ImageRegistrationMethod()
    # Similarity metric settings.
    R.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
    R.SetMetricSamplingStrategy(R.RANDOM)
    R.SetMetricSamplingPercentage(0.01)

    R.SetInterpolator(sitk.sitkLinear)

    # Optimizer settings.
    if not gradient_descent:
        if nb_dim == 2:
            R.SetOptimizerAsExhaustive([sample_per_axis, sample_per_axis, 0, 0])
            R.SetOptimizerScales(
                [2.0 * pi / sample_per_axis, 2.0 * pi / sample_per_axis, 1.0, 1.0]
            )
        else:
            R.SetOptimizerAsExhaustive(
                [sample_per_axis, sample_per_axis, sample_per_axis, 0, 0, 0]
            )
            R.SetOptimizerScales(
                [
                    2.0 * pi / sample_per_axis,
                    2.0 * pi / sample_per_axis,
                    2.0 * pi / sample_per_axis,
                    1.0,
                    1.0,
                    1.0,
                ]
            )
    if gradient_descent:
        R.SetOptimizerAsGradientDescent(0.1, 100)
    R.SetInitialTransform(initial_transform, inPlace=False)

    # Connect all of the observers so that we can perform plotting during registration.
    R.SetGlobalDefaultNumberOfThreads(threads)
    final_transform = R.Execute(
        sitk.Cast(fixed_image, sitk.sitkFloat32),
        sitk.Cast(moving_image, sitk.sitkFloat32),
    )

    moving_resampled = sitk.Resample(
        moving_image,
        fixed_image,
        final_transform,
        sitk.sitkLinear,
        0.0,
        moving_image.GetPixelID(),
    )
    moving_resampled = sitk.GetArrayFromImage(moving_resampled)
    angle_X, angle_Y, angle_Z, _, _, _ = final_transform.GetParameters()
    return 180 * np.array([angle_X, angle_Y, angle_Z]) / np.pi, moving_resampled


def shift_registration_exhaustive_search(
    im1, im2, t_min=-20, t_max=20, t_step=4, fourier_space=False
):
    if fourier_space:
        ft1 = im1
        ft2 = im2
    else:
        ft1 = np.fft.fftn(im1)
        ft2 = np.fft.fftn(im2)
    trans_vecs = np.arange(t_min, t_max, t_step)
    grid_trans_vec = np.array(
        np.meshgrid(trans_vecs, trans_vecs, trans_vecs)
    ).T.reshape((len(trans_vecs) ** 3, 3))
    # print('len', len(grid_trans_vec))
    min_err = 10**20
    best_i = 0
    for i, trans_vec in enumerate(grid_trans_vec):
        ft2_shifted = fourier_shift(ft2, trans_vec)
        err = np.linalg.norm(ft2_shifted - ft1)
        if err < min_err:
            best_i = i
            min_err = err
    res = fourier_shift(ft2, grid_trans_vec[best_i])
    if not fourier_space:
        res = np.fft.ifftn(res)
    return grid_trans_vec[best_i], res.real.astype(im2.dtype)


def center_connected_component(image):
    # Step 1: Find connected components
    labeled_image, num_features = label(image > threshold_otsu(image))

    if num_features < 2:
        return image, np.array([0.0, 0.0, 0.0])  # Already has only one component

    # Step 2: Identify the largest connected component
    component_props = regionprops(labeled_image)
    largest_component = max(component_props, key=lambda prop: prop.area)

    # Step 3: Calculate translation
    center_of_mass = largest_component.centroid
    shift = np.asarray(image.shape) / 2 - np.asarray(center_of_mass)
    shift = np.rint(shift).astype(int)

    # Step 4: Apply translation
    translated_image = np.roll(image, tuple(shift), axis=tuple(range(image.ndim)))

    return translated_image, shift.astype(float)



import numpy as np
import pytest
from scipy.spatial.transform import Rotation as R

import spfluo.data as data
from spfluo.tests.helpers import assert_volumes_aligned
from spfluo.utils.rotate_symmetry_axis import (
    find_pose_from_z_axis_centered_to_centriole_axis,
)
from spfluo.utils.transform import (
    compose_poses,
    distance_poses,
    get_transform_matrix_from_pose,
    invert_pose,
    symmetrize_poses,
)
from spfluo.utils.volume import (
    affine_transform,
)

EPS = 5  # angle error margin in degrees


def distance_transformations(rot1, trans1, rot2, trans2):
    pose1 = np.concatenate((R.from_matrix(rot1).as_euler("XZX", degrees=True), trans1))
    pose2 = np.concatenate((R.from_matrix(rot2).as_euler("XZX", degrees=True), trans2))
    return distance_poses(pose1, pose2, symmetry=9, ignore_symmetry=False)


def distance_rotation_matrices(A, B):
    vec = np.array([1, 0, 0])
    return np.linalg.norm((A - B) @ vec)


@pytest.fixture(scope="module")
def create_data():
    d = data.generated_anisotropic()
    volumes, poses, gt, _ = d["volumes"], d["poses"], d["gt"], d["psf"]
    pose = np.asarray([45, 45, 0, 5, -5, 2], dtype=float)
    gt_rotated = affine_transform(
        gt, np.linalg.inv(get_transform_matrix_from_pose(volumes[0].shape, pose))
    )
    poses_rotated = compose_poses(invert_pose(pose), poses)
    return (volumes, poses_rotated, gt_rotated), (volumes, poses, gt)


def test_data(create_data, save_result):
    (volumes, poses, reconstruction), _ = create_data
    # Vérification que les données sont bonnes
    volumes_rotated_back = affine_transform(
        volumes, get_transform_matrix_from_pose(volumes[0].shape, poses), batch=True
    )
    save_result("volumes_rotated_back", volumes_rotated_back)
    save_result("gt_rotated", reconstruction)
    assert_volumes_aligned(reconstruction, volumes_rotated_back, atol=1)


def test_find_rot_mat_easy(create_data):
    _, (_, _, centriole) = create_data
    pose = find_pose_from_z_axis_centered_to_centriole_axis(centriole, 9)
    assert np.isclose(
        distance_rotation_matrices(
            R.from_euler("XZX", pose[:3], degrees=True).as_matrix(), np.identity(3)
        ),
        0,
        atol=1e-1,
    )


def test_apply_transformation_and_sym_to_poses(create_data, save_result):
    (
        (volumes, poses, reconstruction),
        (_, poses_true_aligned, reconstruction_true_aligned),
    ) = create_data

    pose_from_axis_to_reconstruction = find_pose_from_z_axis_centered_to_centriole_axis(
        reconstruction, 9, center_precision=0.5
    )
    reconstruction_aligned = affine_transform(
        reconstruction,
        get_transform_matrix_from_pose(
            reconstruction.shape, pose_from_axis_to_reconstruction
        ),
    )
    save_result("reconstruction", reconstruction)
    save_result("reconstruction_true_aligned", reconstruction_true_aligned)
    save_result("reconstruction_aligned", reconstruction_aligned)
    assert_volumes_aligned(reconstruction_aligned, reconstruction_true_aligned, atol=1)

    # Test poses_aligned are ok
    poses_from_reconstruction_to_vols = poses
    poses_aligned = compose_poses(
        pose_from_axis_to_reconstruction, poses_from_reconstruction_to_vols
    )
    volumes_aligned = affine_transform(
        volumes,
        get_transform_matrix_from_pose(volumes[0].shape, poses_aligned),
        batch=True,
    )
    save_result("volumes_aligned", volumes_aligned)
    assert_volumes_aligned(volumes_aligned, reconstruction_true_aligned, atol=1)
    angular_errors, trans_errors = distance_poses(
        poses_aligned, poses_true_aligned, symmetry=9
    )
    bottom_up_pose = np.asarray(
        [0, 180, 0, 0, 0, 0]
    )  # alignement can put the image upside down
    poses_aligned_flipped = compose_poses(bottom_up_pose, poses_aligned)
    volumes_aligned = affine_transform(
        volumes,
        get_transform_matrix_from_pose(volumes[0].shape, poses_aligned_flipped),
        batch=True,
    )
    save_result("volumes_aligned_flipped", volumes_aligned)
    angular_errors_flipped, _ = distance_poses(
        poses_aligned_flipped, poses_true_aligned, symmetry=9
    )
    assert (
        all(np.isclose(angular_errors, 0, atol=360 / 9 / 2 + EPS))
        or all(np.isclose(angular_errors_flipped, 0, atol=360 / 9 / 2 + EPS))
    ) and all(np.isclose(trans_errors, 0, atol=1.5))

    # Test symmetrizing these poses
    poses_aligned_sym = symmetrize_poses(poses_aligned, 9)
    volume0_aligned_sym = affine_transform(
        np.stack((volumes[0],) * 9),
        get_transform_matrix_from_pose(volumes[0].shape, poses_aligned_sym[0]),
        batch=True,
    )
    save_result("volume0_aligned_sym", volume0_aligned_sym, metadata={"axes": "TZYX"})

    for i in range(volumes.shape[0]):
        volume_i_aligned_sym = affine_transform(
            np.stack((volumes[i],) * 9),
            get_transform_matrix_from_pose(volumes[0].shape, poses_aligned_sym[i]),
            batch=True,
        )
        assert_volumes_aligned(
            volume_i_aligned_sym, reconstruction_true_aligned, atol=1
        )


@pytest.mark.skip(reason="to fix")
def test_real_data(save_result):
    d = data.real_ab_initio_reconstruction()
    image, pose = d["reconstruction"], d["pose"]

    aligned_image = affine_transform(
        image, get_transform_matrix_from_pose(image.shape[:3], pose)
    )

    pose_estimate = find_pose_from_z_axis_centered_to_centriole_axis(
        image, 9, center_precision=1
    )

    bottom_up_pose = np.asarray(
        [0, 180, 0, 0, 0, 0]
    )  # alignement can put the image upside down
    pose_estimate = compose_poses(bottom_up_pose, pose_estimate)
    aligned_image2 = affine_transform(
        image, get_transform_matrix_from_pose(image.shape[:3], pose_estimate)
    )

    save_result("image", image)
    save_result("aligned_image", aligned_image)
    save_result("aligned_image_estimate", aligned_image2)

    angular_distance, trans_distance = distance_poses(pose, pose_estimate, symmetry=9)

    print(pose, pose_estimate)
    np.testing.assert_allclose(angular_distance, 0, atol=360 / 9 / 2 + EPS)
    np.testing.assert_allclose(trans_distance, 0, atol=2)



from .refinement import (
    convolution_matching_poses,
    find_angles_grid,
    reconstruction_L2,
    refine,
)

__all__ = [
    convolution_matching_poses,
    find_angles_grid,
    reconstruction_L2,
    refine,
]



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from scipy.interpolate import griddata


def plot_map(
    vals_to_plot: np.ndarray,
    thetas: np.ndarray,
    phis: np.ndarray,
    vmin: float | None = None,
    vmax: float | None = None,
    ax: Axes | None = None,
):
    if ax is None:
        ax: Axes = plt.subplots()[1]
    min_theta, max_theta, min_phi, max_phi = (
        np.min(thetas),
        np.max(thetas),
        np.min(phis),
        np.max(phis),
    )
    # min_theta, max_theta, min_phi, max_phi = 0,360,0,180
    grid_theta, grid_phi = np.mgrid[min_theta:max_theta:500j, min_phi:max_phi:500j]
    grid_z = griddata(
        np.array([thetas, phis]).T,
        vals_to_plot,
        (grid_theta, grid_phi),
        method="nearest",
    )
    mean_non_nan = np.nanmean(grid_z)
    ax.imshow(
        np.nan_to_num(grid_z.T, nan=mean_non_nan),
        extent=(min_theta, max_theta, min_phi, max_phi),
        origin="lower",
        vmin=vmin,
        vmax=vmax,
    )  # , origin='lower', extent=(0, 1, 0, 1))
    ax.set_xlabel("φ_1 (°)")
    ax.set_ylabel("φ_2 (°)")
    return ax



import atexit
import os
import tempfile
from typing import List, Tuple

import napari
import numpy as np
import tifffile

from spfluo.manual_picking.annotate import annotate


def show_points(
    im_path: str, csv_path: str, scale: tuple[float, float, float] | None = None
):
    scale = scale if scale else (1, 1, 1)
    annotate(im_path, csv_path, spacing=scale, save=False)


def show_particles(im_paths: List[str], scale: Tuple[float, float, float] = None):
    viewer = napari.Viewer()

    f = tempfile.NamedTemporaryFile(suffix=".tif", delete=False)
    f.close()
    atexit.register(lambda: os.remove(f.name))
    ims = np.stack([tifffile.imread(p) for p in im_paths])
    tifffile.imwrite(f.name, ims)
    (layer,) = viewer.open(f.name, colormap="gray", name="particle")
    if scale:
        layer.scale = scale
    napari.run()



import tempfile
from pathlib import Path

from spfluo import data
from spfluo.refinement.__main__ import create_parser, main


def test_main_refinement():
    generated_root_dir = data.generated_anisotropic()["rootdir"]
    parser = create_parser()
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = Path(tmpdirname)
        reconstruction_path = tmpdir / "reconstruction.tiff"
        poses_path = tmpdir / "poses.csv"
        args = parser.parse_args(
            [
                "--particles_dir",
                str(generated_root_dir / "particles"),
                "--psf_path",
                str(generated_root_dir / "psf.tiff"),
                "--guessed_poses_path",
                str(generated_root_dir / "poses.csv"),
                "--ranges",
                "0",
                "10",
                "--steps",
                "(1,1)",
                "2",
                "--output_reconstruction_path",
                str(reconstruction_path),
                "--output_poses_path",
                str(poses_path),
            ]
        )
        main(args)
        assert reconstruction_path.exists()
        assert poses_path.exists()



"""The pretraining submodule aims at generating datasets, and preparing data for training
(crop + split) regardless of the task. The other two submodules (training and postraining)
however  are specific to the picking task. That is why everything related to tilt
(data loading, training, prediction) is contained separately here within a single file.
"""

import os
import pickle
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset

from ..utils import load_annotations, load_array, summary
from .augment import get_augment_policy
from .core import Accuracy, Data, Logger, Losses, fit, make_model
from .prepare_fs_data import get_patch_paths
from .unsupervised_pretraining import unsupervised_pretraining

# +------------------------------------------------------------------------------------------+ #
# |                                          DATASET                                         | #
# +------------------------------------------------------------------------------------------+ #


class Trainset(Dataset):
    def __init__(
        self,
        paths: List[str],
        labels: np.ndarray,
        patch_size: Tuple[int],
        extension: str,
        augment: float,
        training: bool,
        size: int = None,
    ) -> None:
        super().__init__()
        self.paths = paths
        self.labels = labels
        self.extension = extension
        self.augment = get_augment_policy(patch_size, p=augment, dim=2)
        self.training = training
        self.size = size
        self.unsupervised = self.labels is None

    def get_target(self, path: str) -> int:
        filename = os.path.basename(path)
        if self.extension == "npz":
            y = filename.split("_")
            image_name, patch_index = "_".join(y[:-1]), y[-1]
        else:
            temp = filename.split("_")
            image_name, patch_index = "_".join(temp[:-1]), temp[-1]
        patch_index = int(patch_index.split(".")[0])
        image_name = f"{image_name}.{self.extension}"
        target = self.labels[self.labels[:, 0] == image_name][patch_index][1]
        return torch.as_tensor(target)

    def get_image(self, path: str):
        image = load_array(path)
        image = image.sum(axis=0)  # sum over z axis
        image = (image - image.min()) / (image.max() - image.min())
        if self.training:
            image = self.augment(image=image)["image"]
        image = (image - image.min()) / (image.max() - image.min())
        image = np.expand_dims(image, axis=0)  # add channel dim at the beginning
        return torch.from_numpy(image.astype(np.float32))

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int]:
        if index >= len(self.paths):
            np.random.shuffle(self.paths)
        index = index % len(self.paths)
        target = self.get_target(self.paths[index]) if not self.unsupervised else 0
        return self.get_image(self.paths[index]), target

    def __len__(self) -> int:
        return self.size if self.size is not None else len(self.paths)


def get_patch_size(rootdir: str) -> Tuple[int]:
    train_paths = get_patch_paths(os.path.join(rootdir, "train", "cropped"))
    return load_array(train_paths[0]).shape


# +------------------------------------------------------------------------------------------+ #
# |                                    TRAINING STRUCTURES                                   | #
# +------------------------------------------------------------------------------------------+ #


def make_data(
    rootdir: str,
    output_dir: str,
    extension: str,
    batch_size: int,
    num_workers: int,
    shuffle: bool,
    augment: float,
    size: int,
    patch_size: Tuple[int],
    num_pos_samples: int,
    unsupervised: bool,
    selected_particles: str,
    crop_dir: str,
) -> Tuple[DataLoader]:
    # labels
    if unsupervised:
        train_views, val_views = None, None
    else:
        if extension == "npz":
            views = load_annotations(os.path.join(rootdir, "views.csv"))
            train_views = val_views = views
        else:
            train_views = load_annotations(
                os.path.join(rootdir, "train", "train_views.csv")
            )
            val_views = load_annotations(os.path.join(rootdir, "val", "val_views.csv"))

    if crop_dir is None:
        datadir = os.path.join(rootdir, "train", "cropped")
    else:
        datadir = crop_dir

    # train data
    if selected_particles is None:
        train_paths = get_patch_paths(datadir, num_pos_samples, shuffle)
        train_paths = sorted(list(filter(lambda x: "positive" in x, train_paths)))
    else:
        with open(selected_particles, "rb") as f:
            picked_particles = pickle.load(f)
        width = int(
            np.ceil(np.log10(len(os.listdir(os.path.join(datadir, "positive")))))
        )
        train_paths = list(
            map(
                lambda x: os.path.join(
                    datadir,
                    "positive",
                    x[1].replace(
                        "." + extension, "_" + str(x[0]).zfill(width) + ".npz"
                    ),
                ),
                picked_particles,
            )
        )
    with open(os.path.join(output_dir, "selected_particles.pickle"), "wb") as file:
        pickle.dump(train_paths, file)
    args = [train_paths, train_views, patch_size, extension, augment]
    # print(train_paths)
    train_set = Trainset(*args, training=True, size=size)
    # val data
    if crop_dir is None:
        datadir = os.path.join(rootdir, "val", "cropped")
    else:
        datadir = crop_dir
    val_paths = get_patch_paths(datadir)  # num_pos_samples=None, shuffle=False
    val_paths = sorted(list(filter(lambda x: "positive" in x, val_paths)))
    val_set = Trainset(
        val_paths, val_views, patch_size, extension, augment=0, training=False
    )
    # loaders
    loader_kwargs = {"batch_size": batch_size, "num_workers": num_workers}
    train_loader = DataLoader(train_set, **loader_kwargs, shuffle=True)
    val_loader = DataLoader(val_set, **loader_kwargs, shuffle=False)
    return Data(train_loader, val_loader)


def make_logger(config: Dict) -> Logger:
    # Names order MUST match MultiTaskLoss outputs order.
    loss_names = ["classification"]
    if config["ae"]:
        loss_names.append("reconstruction")
    loss_names.append("total")
    losses = Losses(loss_names)
    accuracy = Accuracy()
    return Logger(losses, accuracy, config)


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def train(
    rootdir: str,
    batch_size: int,
    num_workers: int,
    output_dir: str,
    extension: str,
    selected_particles: str,
    crop_dir: str,
    num_epochs: int,
    lr: float,
    network: str,
    width_factor: float,
    depth_factor: float,
    swa: bool,
    swa_lr: float,
    patch_size: Tuple[int],
    num_pos_samples: int,
    shuffle: bool,
    augment: float,
    size: int = None,
    # autoencoder params
    ae: bool = False,
    eta: float = 0.0,
    learn_eta: bool = False,
    num_features: Tuple[int] = None,
    hidden_dim: int = None,
    # unsupervised pretraining
    pretrain: bool = False,
) -> None:
    print("\nPreparing training pipeline ...\n")
    config = locals()
    os.makedirs(output_dir, exist_ok=True)
    summary_output = os.path.join(output_dir, "tilt_train_config.txt")
    summary(config, title="TILT TRAINING CONFIG", output=summary_output)
    num_classes, dim, pi = 3, 2, None
    data_params = [
        rootdir,
        output_dir,
        extension,
        batch_size,
        num_workers,
        shuffle,
        augment,
        size,
        patch_size,
    ]
    train_params = [num_classes, dim, lr, patch_size, num_epochs, pi]
    model_params = [network, width_factor, depth_factor, swa, swa_lr]
    ae_params = [ae, eta, learn_eta, num_features, hidden_dim]
    if pretrain:
        # autoencoder reconstruction task for unsupervised pretraining of the encoder.
        data = make_data(
            *data_params, None, True, None, crop_dir
        )  # num_pos_samples=None <=> use all samples
        model = make_model(
            *train_params, *model_params, *[True, *ae_params[1:]]
        )  # ae = True
        model.train_criterion = model.val_criterion = torch.nn.MSELoss()
        model = unsupervised_pretraining(data, model, num_epochs, output_dir)
        pretrained_encoder_state = model.network.state_dict()
    data = make_data(*data_params, num_pos_samples, False, selected_particles, None)
    model = make_model(*train_params, *model_params, *ae_params)
    if pretrain:
        model.network.load_state_dict(pretrained_encoder_state)
    logger = make_logger(config)
    fit(data, model, logger)



import random

# DEBUG only flag
VERBOSE = False


class Transform:
    def __init__(self, always_apply=False, p=0.5):
        assert 0 <= p <= 1
        self.p = p
        self.always_apply = always_apply

    def __call__(self, force_apply, targets, **data):
        if force_apply or self.always_apply or random.random() < self.p:
            params = self.get_params(**data)
            if VERBOSE:
                print("RUN", self.__class__.__name__, params)
            for k, v in data.items():
                if k in targets[0]:
                    data[k] = self.apply(v, **params)
                else:
                    data[k] = v
        return data

    def get_params(self, **data):
        """
        shared parameters for one apply. (usually random values)
        """
        return {}

    def apply(self, volume, **params):
        raise NotImplementedError


class DualTransform(Transform):
    def __call__(self, force_apply, targets, **data):
        if force_apply or self.always_apply or random.random() < self.p:
            params = self.get_params(**data)
            if VERBOSE:
                print("RUN", self.__class__.__name__, params)

            for k, v in data.items():
                if k in targets[0]:
                    data[k] = self.apply(v, **params)
                elif k in targets[1]:
                    data[k] = self.apply_to_mask(v, **params)
                else:
                    data[k] = v
        return data

    def apply_to_mask(self, mask, **params):
        return self.apply(mask, **params)



"""
Multiple viewer widget
======================

This is an example on how to have more than one viewer in the same napari window.
Additional viewers state will be synchronized with the main viewer.
Switching to 3D display will only impact the main viewer.

This example also contain option to enable cross that will be moved to the
current dims point (`viewer.dims.point`).

.. tags:: gui
"""

from copy import deepcopy

import napari
import napari_bbox
import numpy as np
from napari.components.layerlist import Extent
from napari.components.viewer_model import ViewerModel
from napari.layers import Image, Labels, Layer, Points, Vectors
from napari.qt import QtViewer
from napari.utils.action_manager import action_manager
from napari.utils.events import Event
from napari.utils.events.event import WarningEmitter
from napari.utils.notifications import show_info
from packaging.version import parse as parse_version
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QDoubleSpinBox,
    QPushButton,
    QSplitter,
    QVBoxLayout,
    QWidget,
)
from superqt.utils import qthrottled

NAPARI_GE_4_16 = parse_version(napari.__version__) > parse_version("0.4.16")


def copy_layer_le_4_16(layer: Layer, name: str = ""):
    res_layer = deepcopy(layer)
    # this deepcopy is not optimal for labels and images layers
    if isinstance(layer, (Image, Labels)):
        res_layer.data = layer.data

    res_layer.metadata["viewer_name"] = name

    res_layer.events.disconnect()
    res_layer.events.source = res_layer
    for emitter in res_layer.events.emitters.values():
        emitter.disconnect()
        emitter.source = res_layer
    return res_layer


def copy_layer(layer: Layer, name: str = ""):
    if not NAPARI_GE_4_16:
        return copy_layer_le_4_16(layer, name)

    res_layer = Layer.create(*layer.as_layer_data_tuple())
    res_layer.metadata["viewer_name"] = name
    return res_layer


def get_property_names(layer: Layer):
    klass = layer.__class__
    res = []
    for event_name, event_emitter in layer.events.emitters.items():
        if isinstance(event_emitter, WarningEmitter):
            continue
        if event_name in ("thumbnail", "name"):
            continue
        if (
            isinstance(getattr(klass, event_name, None), property)
            and getattr(klass, event_name).fset is not None
        ):
            res.append(event_name)
    return res


def center_cross_on_mouse(
    viewer_model: napari.components.viewer_model.ViewerModel,
):
    """move the cross to the mouse position"""

    if not getattr(viewer_model, "mouse_over_canvas", True):
        # There is no way for napari 0.4.15 to check if mouse is over sending canvas.
        show_info("Mouse is not over the canvas. You may need to click on the canvas.")
        return

    viewer_model.dims.current_step = tuple(
        np.round(
            [
                max(min_, min(p, max_)) / step
                for p, (min_, max_, step) in zip(
                    viewer_model.cursor.position, viewer_model.dims.range
                )
            ]
        ).astype(int)
    )


action_manager.register_action(
    name="napari:move_point",
    command=center_cross_on_mouse,
    description="Move dims point to mouse position",
    keymapprovider=ViewerModel,
)

action_manager.bind_shortcut("napari:move_point", "C")


class own_partial:
    """
    Workaround for deepcopy not copying partial functions
    (Qt widgets are not serializable)
    """

    def __init__(self, func, *args, **kwargs) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        return self.func(*(self.args + args), **{**self.kwargs, **kwargs})

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        return own_partial(
            self.func,
            *deepcopy(self.args, memodict),
            **deepcopy(self.kwargs, memodict),
        )


class QtViewerWrap(QtViewer):
    def __init__(self, main_viewer, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.main_viewer = main_viewer

    def _qt_open(
        self,
        filenames: list,
        stack: bool,
        plugin: str = None,
        layer_type: str = None,
        **kwargs,
    ):
        """for drag and drop open files"""
        self.main_viewer.window._qt_viewer._qt_open(
            filenames, stack, plugin, layer_type, **kwargs
        )


class CrossWidget(QCheckBox):
    """
    Widget to control the cross layer. because of the performance reason
    the cross update is throttled
    """

    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__("Add cross layer")
        self.viewer = viewer
        self.setChecked(False)
        self.stateChanged.connect(self._update_cross_visibility)
        self.layer = None
        self.viewer.dims.events.order.connect(self.update_cross)
        self.viewer.dims.events.ndim.connect(self._update_ndim)
        self.viewer.dims.events.current_step.connect(self.update_cross)
        self._extent = None

        self._update_extent()
        self.viewer.dims.events.connect(self._update_extent)

    @qthrottled(leading=False)
    def _update_extent(self):
        """
        Calculate the extent of the data.

        Ignores the the cross layer itself in calculating the extent.
        """
        if NAPARI_GE_4_16:
            layers = [layer for layer in self.viewer.layers if layer is not self.layer]
            self._extent = self.viewer.layers.get_extent(layers)
        else:
            extent_list = [
                layer.extent for layer in self.viewer.layers if layer is not self.layer
            ]
            self._extent = Extent(
                data=None,
                world=self.viewer.layers._get_extent_world(extent_list),
                step=self.viewer.layers._get_step_size(extent_list),
            )
        self.update_cross()

    def _update_ndim(self, event):
        if self.layer in self.viewer.layers:
            self.viewer.layers.remove(self.layer)
        self.layer = Vectors(name=".cross", ndim=event.value)
        self.layer.edge_width = 1.5
        self.update_cross()

    def _update_cross_visibility(self, state):
        if state:
            self.viewer.layers.append(self.layer)
        else:
            self.viewer.layers.remove(self.layer)
        self.update_cross()

    def update_cross(self):
        if self.layer not in self.viewer.layers:
            return

        point = self.viewer.dims.current_step
        vec = []
        for i, (lower, upper) in enumerate(self._extent.world.T):
            if (upper - lower) / self._extent.step[i] == 1:
                continue
            point1 = list(point)
            point1[i] = (lower + self._extent.step[i] / 2) / self._extent.step[i]
            point2 = [0 for _ in point]
            point2[i] = (upper - lower) / self._extent.step[i]
            vec.append((point1, point2))
        if np.any(self.layer.scale != self._extent.step):
            self.layer.scale = self._extent.step
        self.layer.data = vec


class ExampleWidget(QWidget):
    """
    Dummy widget showcasing how to place additional widgets to the right
    of the additional viewers.
    """

    def __init__(self) -> None:
        super().__init__()
        self.btn = QPushButton("Perform action")
        self.spin = QDoubleSpinBox()
        layout = QVBoxLayout()
        layout.addWidget(self.spin)
        layout.addWidget(self.btn)
        layout.addStretch(1)
        self.setLayout(layout)


class MultipleViewerWidget(QSplitter):
    """The main widget of the example."""

    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__()
        self.viewer = viewer
        self.viewer_model1 = ViewerModel(title="model1")
        self.viewer_model2 = ViewerModel(title="model2")
        self._block = False
        self.qt_viewer1 = QtViewerWrap(viewer, self.viewer_model1)
        self.qt_viewer2 = QtViewerWrap(viewer, self.viewer_model2)
        viewer_splitter = QSplitter()
        viewer_splitter.setOrientation(Qt.Vertical)
        viewer_splitter.addWidget(self.qt_viewer1)
        viewer_splitter.addWidget(self.qt_viewer2)
        viewer_splitter.setContentsMargins(0, 0, 0, 0)

        self.addWidget(viewer_splitter)

        self.viewer.layers.events.inserted.connect(self._layer_added)
        self.viewer.layers.events.removed.connect(self._layer_removed)
        self.viewer.layers.events.moved.connect(self._layer_moved)
        self.viewer.layers.selection.events.active.connect(
            self._layer_selection_changed
        )
        self.viewer.dims.events.current_step.connect(self._point_update)
        self.viewer_model1.dims.events.current_step.connect(self._point_update)
        self.viewer_model2.dims.events.current_step.connect(self._point_update)
        self.viewer.dims.events.order.connect(self._order_update)
        self.viewer.events.reset_view.connect(self._reset_view)
        self.viewer_model1.events.status.connect(self._status_update)
        self.viewer_model2.events.status.connect(self._status_update)

    def _status_update(self, event):
        self.viewer.status = event.value

    def _reset_view(self):
        self.viewer_model1.reset_view()
        self.viewer_model2.reset_view()

    def _layer_selection_changed(self, event):
        """
        update of current active layer
        """
        if self._block:
            return

        if event.value is None:
            self.viewer_model1.layers.selection.active = None
            self.viewer_model2.layers.selection.active = None
            return

        self.viewer_model1.layers.selection.active = self.viewer_model1.layers[
            event.value.name
        ]
        self.viewer_model2.layers.selection.active = self.viewer_model2.layers[
            event.value.name
        ]

    def _point_update(self, event):
        if self._block:
            return
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            if len(model.layers) == 0:
                continue
            if model.dims is event.source:
                continue
            if len(self.viewer.layers) != len(model.layers):
                continue
            model.dims.current_step = event.value

    def _order_update(self):
        order = list(self.viewer.dims.order)
        if len(order) <= 2:
            self.viewer_model1.dims.order = order
            self.viewer_model2.dims.order = order
            return

        order[-3:] = order[-2], order[-3], order[-1]
        self.viewer_model1.dims.order = order
        order = list(self.viewer.dims.order)
        order[-3:] = order[-1], order[-2], order[-3]
        self.viewer_model2.dims.order = order

    def _layer_added(self, event: Event):
        """add layer to additional viewers and connect all required events"""
        layer_added: Layer = event.value
        self.viewer_model1.layers.insert(event.index, copy_layer(layer_added, "model1"))
        self.viewer_model2.layers.insert(event.index, copy_layer(layer_added, "model2"))
        for name in get_property_names(layer_added):
            getattr(layer_added.events, name).connect(
                own_partial(self._property_sync, name)
            )

        if (
            isinstance(layer_added, Labels)
            or isinstance(layer_added, Points)
            or isinstance(layer_added, napari_bbox.BoundingBoxLayer)
        ):
            layer_added.events.set_data.connect(self._set_data_refresh)
            self.viewer_model1.layers[layer_added.name].events.set_data.connect(
                self._set_data_refresh
            )
            self.viewer_model2.layers[layer_added.name].events.set_data.connect(
                self._set_data_refresh
            )
        if isinstance(layer_added, Points):
            layer_added.events.current_edge_color.connect(
                self._current_edge_color_refresh
            )
            self.viewer_model1.layers[
                layer_added.name
            ].events.current_edge_color.connect(self._current_edge_color_refresh)
            self.viewer_model2.layers[
                layer_added.name
            ].events.current_edge_color.connect(self._current_edge_color_refresh)
        if layer_added.name != ".cross":
            self.viewer_model1.layers[layer_added.name].events.data.connect(
                self._sync_data
            )
            self.viewer_model2.layers[layer_added.name].events.data.connect(
                self._sync_data
            )

        layer_added.events.name.connect(self._sync_name)

        self._order_update()

    def _sync_name(self, event):
        """sync name of layers"""
        index = self.viewer.layers.index(event.source)
        self.viewer_model1.layers[index].name = event.source.name
        self.viewer_model2.layers[index].name = event.source.name

    def _sync_data(self, event):
        """sync data modification from additional viewers"""
        if self._block:
            return
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            layer = model.layers[event.source.name]
            if layer is event.source:
                continue
            try:
                self._block = True
                layer.data = event.source.data
            finally:
                self._block = False

    def _set_data_refresh(self, event):
        """
        synchronize data refresh between layers
        """
        if self._block:
            return
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            layer = model.layers[event.source.name]
            if layer is event.source:
                continue
            try:
                self._block = True
                layer.refresh()
            finally:
                self._block = False

    def _current_edge_color_refresh(self, event):
        if self._block:
            return
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            layer = model.layers[event.source.name]
            source_layer = event.source
            if layer is source_layer:
                continue
            try:
                self._block = True
                layer.selected_data = event.source.selected_data
            finally:
                self._block = False

    def _layer_removed(self, event):
        """remove layer in all viewers"""
        self.viewer_model1.layers.pop(event.index)
        self.viewer_model2.layers.pop(event.index)

    def _layer_moved(self, event):
        """update order of layers"""
        dest_index = (
            event.new_index if event.new_index < event.index else event.new_index + 1
        )
        self.viewer_model1.layers.move(event.index, dest_index)
        self.viewer_model2.layers.move(event.index, dest_index)

    def _property_sync(self, name, event):
        """Sync layers properties (except the name)"""
        if event.source not in self.viewer.layers:
            return
        try:
            self._block = True
            setattr(
                self.viewer_model1.layers[event.source.name],
                name,
                getattr(event.source, name),
            )
            setattr(
                self.viewer_model2.layers[event.source.name],
                name,
                getattr(event.source, name),
            )
        finally:
            self._block = False


def init_qt():
    from qtpy import QtCore, QtWidgets

    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    # above two lines are needed to allow to undock the widget with
    # additional viewers


def add_orthoviewer_widget(view: napari.Viewer):
    dock_widget = MultipleViewerWidget(view)
    cross = CrossWidget(view)

    view.window.add_dock_widget(dock_widget, name="Sample")
    view.window.add_dock_widget(cross, name="Cross", area="left")
    return view, dock_widget, cross



# This test tests ab-initio + refinement on the anisotropic dataset

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import spfluo
import spfluo.ab_initio_reconstruction.__main__ as ab_initio_main
import spfluo.refinement.__main__ as refinement_main
import spfluo.utils.__main__ as utils_main
from spfluo import data
from spfluo.ab_initio_reconstruction.learning_algorithms.gradient_descent_importance_sampling import (  # noqa: E501
    compute_energy,
)
from spfluo.utils.array import array_namespace, numpy
from spfluo.utils.loading import read_poses
from spfluo.utils.read_save_files import (
    read_image,
    read_images_in_folder,
)
from spfluo.utils.transform import get_transform_matrix
from spfluo.utils.volume import affine_transform, interpolate_to_size

if TYPE_CHECKING:
    from spfluo.utils.array import Array


def get_energy(
    reconstruction: "Array",
    psf: "Array",
    poses: "Array",
    particles: "Array",
):
    xp = array_namespace(reconstruction, psf, poses, particles)
    transforms = get_transform_matrix(
        particles[0].shape, poses[:, :3], poses[:, 3:], degrees=True
    )
    psf = interpolate_to_size(psf, particles[0].shape)
    transforms = xp.asarray(transforms, dtype=particles.dtype)
    particles_rotated = affine_transform(particles, transforms, batch=True)  # inverse
    psf_rotated = affine_transform(
        xp.stack((psf,) * particles.shape[0]), transforms, batch=True
    )

    energy = xp.mean(
        compute_energy(
            xp.fft.fftn(reconstruction),
            xp.fft.fftn(psf_rotated),
            xp.fft.fftn(particles_rotated),
        )
    )
    return energy


# works on a ~12GB VRAM GPU
@pytest.mark.slow
@pytest.mark.skipif(not spfluo.has_torch(), reason="This test requires torch installed")
@pytest.mark.skipif(
    not spfluo.has_torch_cuda(),
    reason="This test requires torch CUDA installed."
    "It would be too long without a GPU.",
)
def test_ab_initio_refinement(tmpdir):
    tmpdir = Path(tmpdir)
    generated_root_dir = Path(data.generated_anisotropic()["rootdir"])
    particles_path = generated_root_dir / "particles"
    psf_path = generated_root_dir / "psf.tiff"

    # Ab initio main
    ab_initio_parser = ab_initio_main.create_parser()
    ab_initio_args = ab_initio_parser.parse_args(
        [
            "--particles_dir",
            str(particles_path),
            "--psf_path",
            str(psf_path),
            "--output_dir",
            str(tmpdir),
            "--gpu",
            # args
            "--N_iter_max",
            str(10),
            "--eps",
            "-10",
            "--lr",
            str(0.2),
        ]
    )
    ab_initio_main.main(ab_initio_args)
    guessed_poses_path = tmpdir / "poses.csv"
    reconstruction_ab_initio_path = tmpdir / "final_recons.tif"
    assert guessed_poses_path.exists()
    assert reconstruction_ab_initio_path.exists()

    # Symmetry axis aligned with the X-axis
    poses_aligned = tmpdir / "poses_aligned.csv"
    recon_aligned = tmpdir / "recons_aligned.tiff"
    utils_parser = utils_main.create_parser()
    utils_args = utils_parser.parse_args(
        [
            "-f",
            "rotate_symmetry_axis",
            "-i",
            str(reconstruction_ab_initio_path),
            "-o",
            str(poses_aligned),
            "--poses",
            str(guessed_poses_path),
            "--rotated-volume",
            str(recon_aligned),
            "--symmetry",
            str(9),
        ]
    )
    utils_main.main(utils_args)

    # Refinement main
    refinement_parser = refinement_main.create_parser()
    reconstruction_refined_path = tmpdir / "reconstruction_refined.tiff"
    poses_refined_path = tmpdir / "poses_refined.csv"
    refinement_args = refinement_parser.parse_args(
        [
            "--particles_dir",
            str(particles_path),
            "--psf_path",
            str(psf_path),
            "--guessed_poses_path",
            str(poses_aligned),
            "--output_reconstruction_path",
            str(reconstruction_refined_path),
            "--output_poses_path",
            str(poses_refined_path),
            # refinement args
            "--symmetry",
            str(9),
            "--steps",
            "(1024,10)",
            "5",
            "5",
            "5",
            "5",
            "--ranges",
            "0",
            "40",
            "20",
            "10",
            "5",
            "-l",
            "0.001",
            "--gpu",
            "--minibatch_size",
            "256",
        ]
    )
    refinement_main.main(refinement_args)
    assert reconstruction_refined_path.exists()
    assert poses_refined_path.exists()

    # Has the energy decreased ?
    particles, _ = read_images_in_folder(str(particles_path), xp=numpy)
    psf = read_image(str(psf_path), xp=numpy)

    # Compute ab initio energy
    reconstruction_ab_initio = read_image(str(reconstruction_ab_initio_path), xp=numpy)
    poses_ab_initio, _ = read_poses(str(guessed_poses_path))
    energy_ab_initio = get_energy(
        reconstruction_ab_initio, psf, poses_ab_initio, particles
    )

    # Compute refinement energy
    reconstruction_refined = read_image(str(reconstruction_refined_path), xp=numpy)
    poses_refined, _ = read_poses(str(poses_refined_path))
    energy_refined = get_energy(reconstruction_refined, psf, poses_refined, particles)

    assert energy_ab_initio > energy_refined



"""This file defines several routines required for positive-unlabelled learning.
It is organised as follows:
1. Preprocessing: create positive and unlabelled masks.
2. Sampling: sample batches with respect to positive ratio in the dataset.
3. Dataset and Dataloaders: Pytorch datastructures with respect to PU principles.
"""

import os
import pickle
from typing import List, Tuple

import numpy as np
import tifffile
import torch
from scipy import ndimage
from skimage import exposure
from skimage.measure import label, regionprops
from sklearn.mixture import GaussianMixture
from torch.utils.data import DataLoader, Dataset, Sampler
from tqdm import tqdm

from ..utils import load_annotations, load_array
from .augment import get_augment_policy

# +------------------------------------------------------------------------------------------+ #
# |                    PREPROCESSING: POSITIVE & UNLABELLED MASKS CREATION                   | #
# +------------------------------------------------------------------------------------------+ #


def find_max_positive_ratio_slice(image: np.ndarray) -> np.ndarray:
    positive_ratios = image.sum(axis=(1, 2)) / np.prod(image.shape[1:])
    return image[np.argmax(positive_ratios)]


def remove_small_objects_by_area(mask: np.ndarray, area_min: int = 50) -> np.ndarray:
    new_mask = np.copy(mask).astype(int)
    labels = label(new_mask)
    areas = np.bincount(labels.flatten())
    for region in regionprops(labels):
        new_mask[region.slice] = areas[region.label]
    new_mask = new_mask >= area_min
    return new_mask


def get_num_pixels_per_particle(radius: int) -> int:
    size = 2 * radius + 1
    mask = np.zeros((size, size, size))
    coords = np.array(np.ogrid[:size, :size, :size], dtype=object)
    center = np.repeat(radius, 3)
    distance = np.linalg.norm(coords - center)
    mask[distance <= radius] = 1
    return mask.sum(dtype=int)


def make_U_mask(
    image: np.ndarray,
    gamma: float = 1,
    gain: float = 1,
    area_min: int = 50,
    dilation_iter: int = 10,
    return_all_steps: bool = False,
) -> np.ndarray:
    # STEP 1: Gamma correction
    step1 = exposure.adjust_gamma(image, gamma, gain)
    # STEP 2: GMM Fitting & Binarize via GMM means
    # GMM fits on one slice: whitest slice if no slice specified
    fit_slice = find_max_positive_ratio_slice(step1)
    classifier = GaussianMixture(n_components=2)
    classifier.fit(fit_slice.reshape((fit_slice.size, 1)))
    threshold = np.mean(classifier.means_)
    step2 = step1 > threshold
    # STEP 3: Remove Small Objects
    step3 = remove_small_objects_by_area(step2, area_min=area_min)
    # STEP 4: Erosion: make a little 3D cube and erode 1 time with it
    cube = ndimage.generate_binary_structure(rank=3, connectivity=3)
    step4 = ndimage.binary_erosion(step3, cube, iterations=1)
    # STEP 5: Dilation: then dilate n times with it
    step5 = ndimage.binary_dilation(
        step4, cube, iterations=dilation_iter, brute_force=True
    )
    return (step1, step2, step3, step4, step5) if return_all_steps else step5


def recenter_if_needed(
    dim: Tuple[int], shape: Tuple[int], z: int, y: int, x: int
) -> Tuple[int]:
    d, h, w = dim
    D, H, W = shape
    z_min, y_min, x_min = z - d // 2, y - h // 2, x - w // 2
    z_min, y_min, x_min = max(0, z_min), max(0, y_min), max(0, x_min)
    z_max, y_max, x_max = z_min + d, y_min + h, x_min + w
    z_max, y_max, x_max = min(D, z_max), min(H, y_max), min(W, x_max)
    z_min, y_min, x_min = z_max - d, y_max - h, x_max - w
    return x_min, x_max, y_min, y_max, z_min, z_max


def make_P_mask(shape: Tuple[int], centers: np.ndarray, radius: int = 10) -> np.ndarray:
    size = 2 * radius + 1
    dim = np.min([shape, [size, size, size]], axis=0)
    base_mask = np.zeros(dim)
    coords = np.array(np.ogrid[: dim[0], : dim[1], : dim[2]], dtype=object)
    center = dim // 2
    distance = np.linalg.norm(coords - center)
    base_mask[distance <= radius] = 1
    mask = np.zeros(shape)
    for center in centers:
        x_min, x_max, y_min, y_max, z_min, z_max = recenter_if_needed(
            dim, shape, *center
        )
        block = mask[z_min:z_max, y_min:y_max, x_min:x_max]
        mask[z_min:z_max, y_min:y_max, x_min:x_max] = np.logical_or(block, base_mask)
    return mask


def prepare_pu_data(
    rootdir: str,
    output_dir: str,
    csv_name: str,
    extension: str,
    radius: int,
    shuffle: bool,
    num_pos_samples: int = None,
    load_u_masks: bool = False,
    save_selected_particles: bool = False,
) -> Tuple[np.ndarray]:
    print(f"| Preparing data from {rootdir}")
    # 1. Load images
    print("| STEP [1/3]: Loading images and annotations ...")
    images_names = sorted(
        list(filter(lambda x: x.endswith(extension), os.listdir(rootdir)))
    )
    images_paths = sorted(list(map(lambda x: os.path.join(rootdir, x), images_names)))
    images = [load_array(path) for path in tqdm(images_paths)]
    # 2. Load and sample annotations
    annotations = load_annotations(os.path.join(rootdir, csv_name))
    annotations = annotations[:, [0, 2, 3, 4]]
    annotations[:, [1, 2, 3]] = annotations[:, [1, 2, 3]].astype(int)
    num = np.zeros(annotations.shape[0], dtype=int)
    for image_name in images_names:
        idx_image = annotations[:, 0] == image_name
        num[idx_image] = np.arange(idx_image.sum())
    annotations_range = np.arange(annotations.shape[0])
    if shuffle:
        np.random.shuffle(annotations_range)
    annotations = annotations[annotations_range[:num_pos_samples]]
    if save_selected_particles:
        with open(os.path.join(output_dir, "selected_particles.pickle"), "wb") as file:
            ann_to_save = np.concatenate(
                (num[annotations_range[:num_pos_samples], None], annotations), axis=1
            )
            pickle.dump(ann_to_save, file)
    # 3. Load or Make PU masks
    if load_u_masks:
        print("| STEP [2/3]: Loading U masks ...")
        U_masks_dir = os.path.join(rootdir, "U_masks")
        U_masks = [
            tifffile.imread(os.path.join(U_masks_dir, name))
            for name in tqdm(images_names)
        ]
    else:
        print("| STEP [2/3]: Making U masks ...")
        U_masks = [make_U_mask(image) for image in images]
    print("| STEP [3/3]: Making P masks ...")
    P_masks = [
        make_P_mask(images[i].shape, annotations[annotations[:, 0] == name, 1:], radius)
        for i, name in tqdm(enumerate(images_names), total=len(images_names))
    ]
    return images, U_masks, P_masks


# +------------------------------------------------------------------------------------------+ #
# |                            POSITIVE - UNLABELLED BATCH SAMPLING                          | #
# +------------------------------------------------------------------------------------------+ #


class ShuffledSampler(Sampler):
    def __init__(self, dataset: np.ndarray) -> None:
        self.dataset = dataset
        self.reset_and_shuffle()

    def reset_and_shuffle(self) -> None:
        self.current_index = 0
        np.random.shuffle(self.dataset)

    def __len__(self) -> int:
        return len(self.dataset)

    def __next__(self) -> Tuple[int, int]:
        if self.current_index >= len(self.dataset):
            self.reset_and_shuffle()
        sample = self.dataset[self.current_index]
        self.current_index += 1
        return sample

    def __iter__(self) -> Sampler:
        return self


class PUSampler(Sampler):
    def __init__(
        self,
        P_masks: List[np.ndarray] = [],
        U_masks: List[np.ndarray] = [],
        pi: float = 0.5,
        batch_size: int = 64,
        epoch_size: int = None,
    ) -> None:
        P = self.enumerate_positive_coordinates(P_masks)
        U = self.enumerate_positive_coordinates(U_masks)
        self.P = ShuffledSampler(P)
        self.U = ShuffledSampler(U)
        self.pi = pi
        self.batch_size = batch_size
        self.num_positives_per_batch = int(pi * batch_size)
        if epoch_size is not None:
            self.size = epoch_size * batch_size
        else:
            self.size = len(self.P) + len(self.U)
        self.__reset_state()

    @staticmethod
    def enumerate_positive_coordinates(masks: np.ndarray) -> np.ndarray:
        size = int(sum(mask.sum() for mask in masks))
        positive_coordinates = np.zeros(
            size, dtype=[("image", np.uint32), ("coord", np.uint32)]
        )
        index = 0
        for mask_index, mask in enumerate(masks):
            positive_indices = np.flatnonzero(mask)
            num_positive_indices = len(positive_indices)
            positive_coordinates["image"][index : index + num_positive_indices] = (
                mask_index
            )
            positive_coordinates["coord"][index : index + num_positive_indices] = (
                positive_indices
            )
            index += num_positive_indices
        return positive_coordinates

    def __reset_state(self) -> None:
        self.current_sampler = self.P
        self.current_num_positives = 0
        self.current_index = 0

    def __update_state(self) -> None:
        self.current_index += 1
        if self.current_num_positives < self.num_positives_per_batch:
            self.current_num_positives += 1
        elif self.current_num_positives == self.num_positives_per_batch:
            self.current_sampler = self.U

    def __len__(self) -> int:
        return self.size

    @staticmethod
    def encode_indices_to_integer(mask_index: int, coord: int) -> np.int64:
        return mask_index * 2**32 + coord

    def __next__(self) -> int:
        self.__update_state()
        mask_index, coord = next(self.current_sampler)
        if self.current_index == self.batch_size:
            self.__reset_state()
        return self.encode_indices_to_integer(mask_index, coord)

    def __iter__(self) -> int:
        for _ in range(len(self)):
            yield next(self)


def estimate_pi(
    U_masks: List[np.ndarray], expected_num_particles_per_image: int, radius: int
) -> float:
    num_images = len(U_masks)
    num_pixels_per_particle = get_num_pixels_per_particle(radius)
    expected_P_size = (
        num_pixels_per_particle * expected_num_particles_per_image * num_images
    )
    U_size = int(sum(mask.sum() for mask in U_masks))
    estimated_pi = expected_P_size / U_size
    return estimated_pi


# +------------------------------------------------------------------------------------------+ #
# |                                   DATASETS & DATALOADERS                                 | #
# +------------------------------------------------------------------------------------------+ #


class Trainset(Dataset):
    def __init__(
        self,
        images: Tuple[np.ndarray],
        labels: Tuple[np.ndarray],
        patch_size: Tuple[int],
        dim: int,
        augment: float,
        training: bool,
    ) -> None:
        super().__init__()
        self.images = images
        self.labels = labels
        self.patch_size = patch_size
        self.dim = dim
        self.augment = get_augment_policy(patch_size, p=augment, dim=dim)
        self.training = training

    def __len__(self):
        return int(sum(label.sum() for label in self.labels))

    @staticmethod
    def decode_integer_to_indices(encoding_integer: int) -> Tuple[int]:
        image_index = encoding_integer // 2**32
        coord = encoding_integer - image_index * 2**32
        return image_index, coord

    def get_patch_slices(self, shape, z, y, x) -> Tuple[int]:
        d, h, w = self.patch_size
        D, H, W = shape
        z_min, y_min, x_min = z - d // 2, y - h // 2, x - w // 2
        z_min, y_min, x_min = max(0, z_min), max(0, y_min), max(0, x_min)
        z_max, y_max, x_max = z_min + d, y_min + h, x_min + w
        z_max, y_max, x_max = min(D, z_max), min(H, y_max), min(W, x_max)
        z_min, y_min, x_min = z_max - d, y_max - h, x_max - w
        return slice(z_min, z_max), slice(y_min, y_max), slice(x_min, x_max)

    def get_patch(self, image_index: int, coord: int) -> np.ndarray:
        image = self.images[image_index]
        d, h, w = image.shape
        z = coord // (h * w)
        xy = coord % (h * w)
        y = xy // w
        x = xy % w
        patch = image[self.get_patch_slices(image.shape, z, y, x)]
        if self.dim == 2:
            patch = patch.sum(axis=0)
            mini, maxi = patch.min(), patch.max()
            if mini != maxi:
                patch = (patch - mini) / (maxi - mini)
            else:
                patch = patch - mini
        if self.training:
            patch = self.augment(image=patch)["image"]
        patch = (patch - patch.min()) / (patch.max() - patch.min())
        return patch

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, int]:
        image_index, coord = self.decode_integer_to_indices(idx)
        patch = self.get_patch(image_index, coord)
        label = self.labels[image_index].ravel()[coord]
        return patch, label


def collate_fn(batch):
    """Shuffle samples and put them into tensors."""
    patches = np.array([sample[0] for sample in batch])
    targets = np.array([sample[1] for sample in batch])
    indices = np.arange(len(batch))
    np.random.shuffle(indices)
    patches = torch.from_numpy(patches[indices])
    targets = torch.from_numpy(targets[indices])
    inputs = torch.unsqueeze(patches, dim=1).float()
    targets = torch.unsqueeze(targets, dim=1).float()
    return inputs, targets


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def make_dataloaders(
    rootdir: str,
    output_dir: str,
    batch_size: int,
    num_workers: int,
    num_pos_samples: int,
    shuffle: bool,
    augment: float,
    epoch_size: int,
    dim: int,
    extension: str,
    radius: int,
    num_particles_per_image: int,
    patch_size: Tuple[int],
    load_u_masks: bool,
) -> Tuple[torch.utils.data.DataLoader]:
    # 1. Load train images and make train PU masks
    datadir, csv_name = os.path.join(rootdir, "train"), "train_coordinates.csv"
    args = [
        datadir,
        output_dir,
        csv_name,
        extension,
        radius,
        shuffle,
        num_pos_samples,
        load_u_masks,
        True,
    ]
    train_images, train_U_masks, train_P_masks = prepare_pu_data(*args)
    # 2. Load val images and make val PU masks
    datadir, csv_name = os.path.join(rootdir, "val"), "val_coordinates.csv"
    args = [
        datadir,
        output_dir,
        csv_name,
        extension,
        radius,
        False,
        None,
        load_u_masks,
        False,
    ]
    val_images, val_U_masks, val_P_masks = prepare_pu_data(*args)
    print(len(val_images), len(val_P_masks), len(val_U_masks))
    print(len(train_images), len(train_P_masks), len(train_U_masks))
    # 3. Make train and val sets
    train_set = Trainset(
        train_images, train_P_masks, patch_size, dim, augment, training=True
    )
    val_set = Trainset(
        val_images, val_P_masks, patch_size, dim, augment, training=False
    )
    # 4. Estimate pi
    pi = estimate_pi([*train_U_masks, *val_U_masks], num_particles_per_image, radius)
    # 5. Make train and val samplers
    kwargs = {"pi": pi, "batch_size": batch_size, "epoch_size": epoch_size}
    train_sampler = PUSampler(train_P_masks, train_U_masks, **kwargs)
    val_sampler = PUSampler(val_P_masks, val_U_masks, **kwargs)
    # 6. Make train and val loaders
    kwargs = {
        "batch_size": batch_size,
        "num_workers": num_workers,
        "collate_fn": collate_fn,
    }
    train_loader = DataLoader(train_set, sampler=train_sampler, **kwargs)
    val_loader = DataLoader(val_set, sampler=val_sampler, **kwargs)
    return train_loader, val_loader



from .networks import EfficientNet
from .picking import train as train_picking
from .prepare_pu_data import make_U_mask
from .tilt import train as train_tilt

__all__ = [
    EfficientNet,
    make_U_mask,
    train_picking,
    train_tilt,
]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import imageio
import numpy as np
import tifffile

from spfluo.utils.array import array_namespace, get_prefered_namespace_device, numpy

if TYPE_CHECKING:
    from spfluo.utils.array import Array, Device, array_api_module


def read_image(
    path,
    dtype: str | None = None,
    xp: "array_api_module | None" = None,  # type: ignore
    device: "Device | None" = None,
    gpu: bool | None = None,
):
    xp, device = get_prefered_namespace_device(xp, device, gpu)
    arr = numpy.asarray(
        imageio.mimread(path, memtest=False),
        dtype=getattr(numpy, dtype) if dtype else None,
    )
    return xp.asarray(arr, device=device)


def read_images_in_folder(
    folder,
    alphabetic_order=True,
    dtype: str | None = None,
    xp: "array_api_module | None" = None,  # type: ignore
    device: "Device | None" = None,
    gpu: bool | None = None,
) -> "Array":
    """read all the images inside folder fold"""
    files = os.listdir(folder)
    if alphabetic_order:
        files = sorted(files)
    images = []
    for fn in files:
        pth = f"{folder}/{fn}"
        im = read_image(pth, dtype, xp, device, gpu)
        images.append(im)
    xp = array_namespace(im)
    return xp.stack(images), files


def save_image(path: str, array: "Array", order: str = None, metadata: dict = None):
    assert path.endswith(".ome.tiff")
    metadata = {} if metadata is None else metadata
    if order is not None:
        metadata.update({"axes": order})
    tifffile.imwrite(path, np.asarray(array, dtype=np.float32), metadata=metadata)


def make_dir(dir):
    """creates folder at location dir if i doesn't already exist"""
    if not os.path.exists(dir):
        print(f"directory {dir} created")
        os.makedirs(dir)


def write_array_csv(
    arr: "Array", path: str, sep: str = ",", names: list[str] | None = None
):
    assert arr.ndim <= 2
    assert len(sep) == 1
    if arr.ndim == 1:
        arr = arr[:, None]
    with open(path, "w") as f:
        if names:
            assert len(names) == arr.shape[1]
            for name in names:
                f.write(name)
                f.write(sep)
            f.seek(f.tell() - 1)
            f.write("\n")
        for line in arr:
            for c in line:
                f.write(str(c))
                f.write(sep)
            f.seek(f.tell() - 1)
            f.write("\n")



"""Some functions from this file were translated from code written by Denis Fortun"""

import itertools
import logging
import math
import time
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

import spfluo.utils.debug as debug
from spfluo.utils.array import array_namespace, get_device, to_device, to_numpy
from spfluo.utils.memory import split_batch
from spfluo.utils.transform import (
    get_transform_matrix,
    get_transform_matrix_from_pose,
    symmetrize_poses,
)
from spfluo.utils.volume import (
    affine_transform,
    discretize_sphere_uniformly,
    interpolate_to_size,
    pad,
    phase_cross_correlation,
)

if TYPE_CHECKING:
    from spfluo.utils.array import Array, Device, array_api_module

refinement_logger = logging.getLogger("spfluo.refinement")


def rotate(
    volumes: "Array", poses: "Array", inverse=False, batch=True, multichannel=False
):
    xp = array_namespace(volumes, poses)
    H = get_transform_matrix(
        volumes.shape[-3:],
        poses[..., :3],
        poses[..., 3:],
        convention="XZX",
        degrees=True,
    )
    if not inverse:  # scipy's affine_transform do inverse transform by default
        H = xp.linalg.inv(H)
    H = xp.asarray(H, dtype=volumes.dtype)
    return affine_transform(
        volumes, H, order=1, prefilter=False, batch=batch, multichannel=multichannel
    )


def reconstruction_L2(
    volumes: "Array",
    psf: "Array",
    poses: "Array",
    lambda_: "Array",
    batch: bool = False,
    symmetry: bool = False,
    device: Optional["Device"] = None,
    batch_size: Optional[int] = None,
    multichannel: bool = False,
):
    """Reconstruct a particule from volumes and their poses.
    M reconstructions can be done at once.

    Args:
        volumes (Array): stack of N 3D images of shape (N, (C), D, D, D)
        psf (Array): 3D image of shape ((C), d, h, w)
        poses (Array):
            stack(s) of N poses of shape ((M), (k), N, 6)
            A 'pose' is represented by 6 numbers
                euler angles in the 'zxz' convention in degrees
                a translation vector tz, ty, tx
                you need N poses to describe your volumes
            k is the degree of symmetry. Optional
                If your volumes have a ninefold symmetry, you might want to multiply
                the number of poses by 9 to get a better reconstruction. For that,
                you will need to symmetry axis. It will also increase compute time.
            M is the number of reconstructions you want to do. Optional
                Usefull for computing several reconstructions from the same
                set of volumes
        lambda (Array): regularization parameters of shape ((M),)
        batch (bool): do M reconstructions at once
            poses must be of shape (M, (k), N, 6), the k dim is optional
        symmetry (bool): use a k-degree symmetry
            poses must be of shape ((M), k, N, 6), the M dim is optional
        device (Device): the device to do the computation on.
        batch_size (int or None) : if None, do all the computation at once
        multichannel (bool): choose if multichannel volumes and psf.
            if True, volumes must be of shape (N, C, D, D, D)

    Returns:
        recon (Array):
            reconstruction(s) of shape ((M), (C), D, D, D) or ((M), (C), D+1, D+1, D+1)
    """
    refinement_logger.info("Calling function reconstruction_L2")

    xp = array_namespace(volumes, poses, psf, lambda_)
    host_device = get_device(volumes)
    compute_device = device

    if multichannel:
        N, C, D, H, W = volumes.shape
        c, d, h, w = psf.shape
        assert c == C
    else:
        N, D, H, W = volumes.shape
        d, h, w = psf.shape
        volumes = volumes[:, None, ...]
        psf = psf[None, ...]
        C = 1
    assert D == H == W

    floating_dtype = volumes.dtype
    assert xp.isdtype(floating_dtype, "real floating")
    assert floating_dtype == psf.dtype
    if floating_dtype == xp.float32:
        complex_dtype = xp.complex64
    elif floating_dtype == xp.float64:
        complex_dtype = xp.complex128
    else:
        raise ValueError(
            f"{floating_dtype=} is not supported. "
            "Only float32 and float64 are accepted."
        )

    refinement_logger.debug(f"floating dtype: {floating_dtype}")
    refinement_logger.debug(f"complex dtype: {complex_dtype}")

    refinement_logger.info(
        "Arguments:"
        f" {N} volumes of size {D}x{D}x{D} with C channels"
        f" PSF of size {d}x{h}x{w}"
    )
    if batch:
        assert poses.ndim > 2
        M = poses.shape[0]
        assert lambda_.shape[0] == M
        refinement_logger.info(
            f"Running in batch mode: {M} reconstructions will be done"
        )
    else:
        M = 1
        poses = xp.expand_dims(poses, axis=0)
        lambda_ = xp.reshape(lambda_, (1,))

    if symmetry:
        assert poses.ndim == 4
        k = poses.shape[1]
        refinement_logger.info(f"Symmetry enabled, of degree k={k}")
    else:
        k = 1
        poses = xp.expand_dims(poses, axis=1)  # shape (1, 1, N, 6)

    new_poses = xp.asarray(poses, copy=True)
    if D % 2 == 0:
        # pad by 1 pixel on the right
        volumes = pad(volumes, ((0, 0), (0, 0), (0, 1), (0, 1), (0, 1)))
        assert volumes.shape == (N, C, D + 1, D + 1, D + 1)
        D = D + 1
        new_poses[..., 3:] -= 0.5
        refinement_logger.info(f"Reshaped volumes to odd size {D}x{D}x{D}")
        resize = True
    else:
        resize = False

    psf = interpolate_to_size(psf, (D, D, D), multichannel=True)

    num = xp.zeros((M, C, D, D, D), dtype=complex_dtype, device=host_device)
    den = xp.zeros_like(num, dtype=floating_dtype)

    poses_psf = xp.zeros_like(new_poses)
    poses_psf[..., :3] = new_poses[..., :3]

    # Move data to compute device
    psf = to_device(psf, compute_device)
    for (start1, end1), (start2, end2), (start3, end3), (start4, end4) in split_batch(
        (M, k, N, C), batch_size
    ):
        number_poses = (end1 - start1) * (end2 - start2)
        number_volumes = end3 - start3
        number_channels = end4 - start4

        # Move data to compute device
        poses_minibatch = to_device(
            new_poses[start1:end1, start2:end2, start3:end3, ...], compute_device
        )
        poses_psf_minibatch = to_device(
            poses_psf[start1:end1, start2:end2, start3:end3, ...], compute_device
        )
        volumes_minibatch = to_device(
            volumes[start3:end3, start4:end4, ...], compute_device
        )
        psf_minibatch = to_device(psf[start4:end4, ...], compute_device)

        y = xp.stack(
            (volumes_minibatch,) * number_poses,
        )
        y = xp.reshape(
            rotate(
                xp.reshape(
                    y, (number_poses * number_volumes, number_channels, D, D, D)
                ),
                xp.reshape(poses_minibatch, (number_poses * number_volumes, 6)),
                inverse=True,
                multichannel=True,
            ),
            (end1 - start1, end2 - start2, number_volumes, number_channels, D, D, D),
        )

        H_ = xp.stack((psf_minibatch,) * number_volumes * number_poses)
        H_ = xp.reshape(
            rotate(
                H_,
                xp.reshape(poses_psf_minibatch, (number_poses * number_volumes, 6)),
                inverse=True,
                multichannel=True,
            ),
            (end1 - start1, end2 - start2, end3 - start3, number_channels, D, D, D),
        )

        H_ = xp.fft.fftn(xp.astype(H_, complex_dtype), axes=(-3, -2, -1))

        # Compute numerator
        y = xp.astype(y, complex_dtype)
        y = xp.fft.fftn(xp.fft.fftshift(y, axes=(-3, -2, -1)), axes=(-3, -2, -1))
        y = xp.conj(H_) * y
        num[start1:end1, start4:end4, ...] += to_device(
            xp.sum(y, axis=(1, 2), dtype=complex_dtype), host_device
        )  # reduce symmetry and N dims

        # Compute denominator
        den[start1:end1, start4:end4, ...] += to_device(
            xp.sum(xp.abs(H_) ** 2, axis=(1, 2), dtype=floating_dtype), host_device
        )
        del H_

    num = num / (N * k)
    den = den / (N * k)

    dxyz = xp.zeros((3, 2, 2, 2), device=host_device, dtype=complex_dtype)
    dxyz[0, 0, 0, 0] = 1
    dxyz[0, 1, 0, 0] = -1
    dxyz[1, 0, 0, 0] = 1
    dxyz[1, 0, 1, 0] = -1
    dxyz[2, 0, 0, 0] = 1
    dxyz[2, 0, 0, 1] = -1

    dxyz_padded = pad(dxyz, ((0, 0), *(((D - 1) // 2, (D - 2) // 2),) * 3))
    D_ = xp.fft.fftn(dxyz_padded, axes=(1, 2, 3))
    DtD = xp.sum(
        (xp.abs(D_) ** 2),
        axis=0,
        dtype=floating_dtype,
    )

    den += xp.asarray(lambda_[:, None, None, None, None] * DtD, dtype=xp.float32)

    num = xp.fft.ifftn(num / den, axes=(-3, -2, -1))
    refinement_logger.debug(f"num dtype: {num.dtype}")
    recon = xp.real(num)
    refinement_logger.debug(f"recon 1 dtype: {recon.dtype}")
    recon = xp.where(
        recon > 0, recon, xp.asarray(0.0, device=host_device, dtype=floating_dtype)
    )
    refinement_logger.debug(f"recon 2 end dtype: {recon.dtype}")

    if resize:
        recon = interpolate_to_size(
            recon, (D - 1, D - 1, D - 1), batch=True, multichannel=True
        )
    refinement_logger.debug(f"recon 3 end dtype: {recon.dtype}")

    if not multichannel:
        recon = recon[:, 0, ...]
    if not batch:
        recon = recon[0, ...]

    if refinement_logger.isEnabledFor(logging.DEBUG):
        p = debug.save_image(
            to_numpy(recon),
            debug.DEBUG_DIR_REFINEMENT,
            reconstruction_L2,
            "reconstruction",
            sequence=batch,
            multichannel=True,
        )
        refinement_logger.debug("Saving reconstruction(s) at " + str(p))

    refinement_logger.debug(f"recon end dtype: {recon.dtype}")
    return recon


def convolution_matching_poses(
    reference: "Array",
    volumes: "Array",
    psf: "Array",
    potential_poses: "Array",
    device: Optional["Device"] = None,
    batch_size: int = 1,
):
    """Find the best pose from a list of poses for each volume.
    There can be a different list of pose for each volume.

    Args:
        reference (Array): reference 3D image of shape (C, D, D, D)
        volumes (Array): volumes to match of shape (N, C, D, D, D)
        psf (Array): 3D PSF of shape (C, d, h, w)
        potential_poses (Array): poses to test of shape (N, M, 6)
        device (Device): the device to do the computation on.
        batch_size (int): the batch size
    
    Returns:
        A tuple of 2 arrays representing the best poses for each volume 
        of shape (N, 6) and the dftRegistration error associated
        to each pose (N,).
    """
    xp = array_namespace(volumes, psf, potential_poses)
    host_device = get_device(volumes)
    compute_device = device

    # Shapes
    N1, M, x = potential_poses.shape
    assert x == 6
    N, C, D, H, W = volumes.shape
    assert D == H == W
    assert N == N1
    C1, D1, H, W = reference.shape
    assert C1 == C and D1 == D and H == D and W == D
    C2, _, _, _ = psf.shape
    assert C2 == C

    new_potential_poses = xp.asarray(potential_poses, copy=True)
    if D % 2 == 0:
        new_potential_poses[..., 3:] += 0.5
    else:
        new_potential_poses[..., 3:] += 1.0

    # PSF
    h = xp.fft.fftn(
        xp.fft.fftshift(
            interpolate_to_size(psf, (D, D, D), multichannel=True), axes=(1, 2, 3)
        ),
        axes=(1, 2, 3),
    )

    shifts = xp.empty((N, M, 3), dtype=reference.dtype, device=host_device)
    errors = xp.empty((N, M), dtype=reference.dtype, device=host_device)

    # Move data to compute device
    h = to_device(h, compute_device)
    reference = to_device(reference, compute_device)
    for (start1, end1), (start2, end2) in tqdm(split_batch((N, M), batch_size), total=(N*M)//batch_size, leave=False, desc="convolution matching"):
        minibatch_size = (end1 - start1) * (end2 - start2)
        potential_poses_minibatch = to_device(
            new_potential_poses[start1:end1, start2:end2], compute_device
        )

        # Volumes to Fourier space
        volumes_freq = xp.fft.fftn(
            to_device(volumes[start1:end1], compute_device), axes=(2, 3, 4)
        )

        # Rotate the reference
        reference_minibatch = xp.stack((reference,) * minibatch_size)
        reference_minibatch = xp.reshape(
            rotate(
                reference_minibatch,
                xp.reshape(potential_poses_minibatch, (minibatch_size, 6)),
                multichannel=True,
            ),
            (end1 - start1, end2 - start2, C, D, D, D),
        )
        reference_minibatch = h * xp.fft.fftn(reference_minibatch, axes=(3, 4, 5))

        # Registration
        sh, err, _ = phase_cross_correlation(
            reference_minibatch,
            volumes_freq[:, None],
            nb_spatial_dims=3,
            normalization=None,
            upsample_factor=10,
            space="fourier",
            multichannel=True,
        )
        sh = xp.stack(list(sh), axis=-1)

        errors[start1:end1, start2:end2] = to_device(err, host_device)
        shifts[start1:end1, start2:end2] = to_device(sh, host_device)

        del volumes_freq, reference_minibatch, err, sh

    errors, best_indices = xp.min(errors, axis=1), xp.argmin(errors, axis=1)
    best_poses = potential_poses[xp.arange(N), best_indices]
    best_poses[:, 3:] = -shifts[xp.arange(N), best_indices]
    return best_poses, errors


def find_L(precision: float):
    return math.ceil(((360 / precision) ** 2) / math.pi)


def create_poses_grid(
    xp: "array_api_module", M_axes: int, M_rot: int, symmetry: int = 1, **array_kwargs
):
    (theta, phi, psi), precision = discretize_sphere_uniformly(
        xp, M_axes, M_rot, product=True, symmetry=symmetry, **array_kwargs
    )
    list_angles = xp.stack([theta, phi, psi], axis=-1)
    M = list_angles.shape[0]
    list_translation = xp.zeros((M, 3), **array_kwargs)
    potential_poses = xp.concat([list_angles, list_translation], axis=1)
    return potential_poses, precision


def find_angles_grid(
    xp: "array_api_module",
    reconstruction: "Array",
    patches: "Array",
    psf: "Array",
    precision: int = 10,
):
    L = find_L(precision)
    potential_poses, _ = create_poses_grid(
        xp,
        L,
        1,
        symmetry=1,
        dtype=reconstruction.dtype,
        device=get_device(reconstruction),
    )
    potential_poses = xp.broadcast_to(potential_poses, (patches.shape[0],)+potential_poses.shape)
    best_poses, best_errors = convolution_matching_poses(
        reconstruction, patches, psf, potential_poses
    )

    return best_poses, best_errors


def get_refined_values1D_uniform(
    xp: "array_api_module", loc: float, N: int, range: float, **array_kwargs
):
    return xp.linspace(loc - range / 2, loc + range / 2, N, **array_kwargs)


def get_refined_valuesND(
    xp: "array_api_module",
    locs: List[float],
    N: List[int],
    ranges: List[float],
    **kwargs,
):
    n = len(locs)
    assert n == len(N) == len(ranges)
    values_1d = [
        get_refined_values1D_uniform(xp, locs[i], N[i], range=ranges[i], **kwargs)
        for i in range(n)
    ]

    return xp.asarray(list(itertools.product(*values_1d)))


def create_poses_refined(
    poses: "Array",
    ranges: List[float],
    M: List[int],
):
    xp = array_namespace(poses)
    n_poses = 1
    for x in M:
        n_poses *= x
    potential_poses = xp.stack((xp.asarray(poses, copy=True),) * n_poses, axis=1)
    potential_poses[..., 3:] = 0
    for i in range(poses.shape[0]):
        potential_poses[i, :, :3] = get_refined_valuesND(
            xp, poses[i, :3], M, ranges, dtype=poses.dtype, device=get_device(poses)
        )

    return potential_poses


def refine(
    patches: "Array",
    psf: "Array",
    guessed_poses: "Array",
    steps: List[Union[Tuple[int, int], int]],
    ranges: List[float],
    initial_volume: Optional["Array"] = None,
    lambda_: float = 100.0,
    symmetry: int = 1,
    convention: str = "XZX",
    device: Optional["Device"] = None,
    batch_size: Optional[int] = None,
):
    """
    Args:
        patches:
            shape (N, C, D, D, D)
        psf:
            shape (C, d, h, w)
        symmetry: if greater than 1, adds a symmetry constraint.
            In that case, the symmetry axis must be parallel to the X-axis.
            See get_transformation_matrix function docs for details
            about the convention.
    """
    assert len(steps) == len(ranges), "steps and ranges lists should have equal length"
    assert len(steps) > 0, "length of steps and ranges lists should be at least 1"
    assert symmetry >= 1, "symmetry should be an integer greater or equal to 1"
    assert lambda_ > 0, f"lambda should be greater than 1, found {lambda_}"
    refinement_logger.debug("Calling function refine")
    xp = array_namespace(patches, psf, guessed_poses)
    N, C, D, H, W = patches.shape
    assert D == H == W
    host_device = get_device(patches)
    compute_device = device
    array_kwargs = dict(dtype=patches.dtype, device=host_device)
    lambda_ = xp.asarray(lambda_, **array_kwargs)

    if initial_volume is not None:
        initial_volume = interpolate_to_size(
            initial_volume, (D, D, D), multichannel=True
        )
        current_reconstruction = initial_volume
    else:
        guessed_poses_sym = symmetrize_poses(
            guessed_poses, symmetry=symmetry, convention=convention
        )
        guessed_poses_sym = xp.permute_dims(guessed_poses_sym, (1, 0, 2))
        initial_reconstruction = reconstruction_L2(
            patches,
            psf,
            guessed_poses_sym,
            lambda_,
            symmetry=True,
            device=compute_device,
            batch_size=batch_size,
            multichannel=True,
        )
        initial_reconstruction = interpolate_to_size(
            initial_reconstruction, (D, D, D), multichannel=True
        )

        current_reconstruction = initial_reconstruction

    current_poses = guessed_poses

    if refinement_logger.isEnabledFor(logging.DEBUG):
        im = to_numpy(current_reconstruction)
        p = debug.save_image(
            im,
            debug.DEBUG_DIR_REFINEMENT,
            refine,
            "initial-reconstruction",
            multichannel=True,
        )
        refinement_logger.debug("Saving current reconstruction at " + str(p))
        all_recons = [im]
        debug.save_image(to_numpy(psf), debug.DEBUG_DIR_REFINEMENT, refine, "psf", multichannel=True)
    
        for j in range(len(current_poses)):
            refinement_logger.debug(
                f"pose[{j}], found: ["
                + ", ".join(
                    [f"{x:.1f}" for x in to_numpy(current_poses[j]).tolist()]
                )
                + "]",
            )
            patch_j_trans = to_numpy(
                affine_transform(
                    patches[j],
                    xp.astype(
                        get_transform_matrix_from_pose((D, D, D), current_poses[j]),
                        patches.dtype,
                    ),
                    multichannel=True,
                    order=1,
                )
            )
            debug.save_image(
                patch_j_trans,
                debug.DEBUG_DIR_REFINEMENT,
                refine,
                f"step0-patch{j}-pose",
                multichannel=True,
            )
    for i in tqdm(range(len(steps)), desc="refine"):
        refinement_logger.debug(f"STEP {i+1}/{len(steps)}")
        t1 = time.time()
        # Poses estimation
        s = steps[i]
        if ranges[i] == 0 and type(s) is tuple:  # Discretization of the whole sphere
            M_axes, M_rot = s
            potential_poses, (precision_axes, precision_rot) = create_poses_grid(
                xp, M_axes, M_rot, symmetry=symmetry, **array_kwargs
            )
            potential_poses = xp.broadcast_to(potential_poses, (N, M_axes*M_rot, 6))
            refinement_logger.debug(
                "[convolution_matching_poses] Searching the whole grid. "
                f"N_axes={M_axes}, N_rot={M_rot}. "
                f"precision_axes={precision_axes:.2f}°, "
                f"precision_rot={precision_rot:.2f}°"
            )
            
        elif isinstance(s, int):  # Refinement around the current poses
            refinement_logger.debug(
                f"[convolution_matching_poses] Refining the poses. range={ranges[i]}, steps={s}"
            )
            t0 = time.time()
            potential_poses = create_poses_refined(current_poses, [ranges[i]] * 3, [s] * 3)
        else:
            raise ValueError(
                "When range==0, steps should be a tuple. "
                "When range>0, steps should be an int. "
                f"Found range={ranges[i]} and steps={s}"
            )

        t0 = time.time()
        current_poses, _ = convolution_matching_poses(
            current_reconstruction,
            patches,
            psf,
            potential_poses,
            device=compute_device,
            batch_size=batch_size,
        )
        refinement_logger.debug(
            f"[convolution_matching_poses] Done in {time.time()-t0:.3f}s"
        )

        # Reconstruction
        refinement_logger.debug("[reconstruction_L2] Reconstruction")
        t0 = time.time()
        current_poses_sym = symmetrize_poses(
            current_poses, symmetry=symmetry, convention=convention
        )
        current_poses_sym = xp.permute_dims(current_poses_sym, (1, 0, 2))
        current_reconstruction = reconstruction_L2(
            patches,
            psf,
            current_poses_sym,
            lambda_,
            symmetry=True,
            device=compute_device,
            batch_size=batch_size,
            multichannel=True,
        )
        current_reconstruction = interpolate_to_size(
            current_reconstruction, (D, D, D), multichannel=True
        )
        refinement_logger.debug(f"[reconstruction_L2] Done in {time.time()-t0:.3f}s")

        if refinement_logger.isEnabledFor(
            logging.DEBUG
        ):  # .cpu() causes host-device sync
            for j in range(len(current_poses)):
                refinement_logger.debug(
                    f"pose[{j}], found: ["
                    + ", ".join(
                        [f"{x:.1f}" for x in to_numpy(current_poses[j]).tolist()]
                    )
                    + "]",
                )
                patch_j_trans = to_numpy(
                    affine_transform(
                        patches[j],
                        xp.astype(
                            get_transform_matrix_from_pose((D, D, D), current_poses[j]),
                            patches.dtype,
                        ),
                        multichannel=True,
                        order=1,
                    )
                )
                debug.save_image(
                    patch_j_trans,
                    debug.DEBUG_DIR_REFINEMENT,
                    refine,
                    f"step{i+1}-patch{j}-pose",
                    multichannel=True,
                )

            im = to_numpy(current_reconstruction)
            p = debug.save_image(
                im, debug.DEBUG_DIR_REFINEMENT, refine, f"step{i+1}", multichannel=True
            )
            refinement_logger.debug("Saving current reconstruction at " + str(p))
            all_recons.append(im)

        refinement_logger.debug(
            f"STEP {i+1}/{len(steps)} done in {time.time()-t1:.3f}s"
        )

    if refinement_logger.isEnabledFor(logging.DEBUG):
        p = debug.save_image(
            np.stack(all_recons, axis=0),
            debug.DEBUG_DIR_REFINEMENT,
            refine,
            "all-steps",
            sequence=True,
            multichannel=True,
        )
        refinement_logger.debug("Saving all reconstructions at " + str(p))

    return current_reconstruction, current_poses



from typing import TYPE_CHECKING

import numpy as np
import scipy.ndimage as scp

from spfluo.utils.array import array_namespace

if TYPE_CHECKING:
    from spfluo.utils.array import Array


def threshold(im, thresh_val):
    im_out = (im >= thresh_val) * im
    return im_out


def crop_center(image, size):
    """crop a sub array of shape 'size' at the center of the image.
    It pads the images at the dimensions where the shape is smaller than
    the crop size"""
    nb_dim = len(image.shape)
    padding = []
    for d in range(nb_dim):
        if image.shape[d] < size[d]:
            padding_d = int((size[d] - image.shape[d]) // 2 + 1)
            padding.append((padding_d, padding_d))
        else:
            padding.append((0, 0))
    padding = tuple(padding)
    image = np.pad(image, padding, mode="constant")
    size = np.array(size)
    c = [image.shape[0] // 2, image.shape[1] // 2, image.shape[2] // 2]
    cropped = image[
        c[0] - size[0] // 2 : c[0] + size[0] // 2,
        c[1] - size[1] // 2 : c[1] + size[1] // 2,
        c[2] - size[2] // 2 : c[2] + size[2] // 2,
    ]
    return cropped


def normalize(arr: "Array", min: float = 0, max: float = 1):
    xp = array_namespace(arr)
    m, M = xp.min(arr), xp.max(arr)
    if M > m:
        norm_0_1 = (arr - m) / (M - m)
    else:
        norm_0_1 = xp.zeros_like(arr)
    norm = (max - min) * norm_0_1 + min
    return norm


def resize(in_array, desired_shape):
    """zoom an image to a desired shape"""
    array_shape = np.array(list(in_array.shape))
    desired_shape = np.array(desired_shape)
    return scp.zoom(in_array, desired_shape / array_shape, order=5)


def resize_to_given_size_of_pixels(volume, pixel_size_before, pixel_size_after):
    pixel_size_before, pixel_size_after = (
        np.array(pixel_size_before),
        np.array(pixel_size_after),
    )
    shape_before = volume.shape
    ratio_shape = pixel_size_before / pixel_size_after
    shape_after = shape_before * ratio_shape
    shape_after = [int(shape_after[d] + 0.5) for d in range(len(pixel_size_after))]
    return resize(volume, shape_after)


def window_fft(im, size, sig_noise):
    fft_im = np.fft.fftshift(np.fft.fftn(im))
    window_fft_im = np.random.normal(0, sig_noise, im.shape)
    c = [im.shape[0] // 2, im.shape[1] // 2, im.shape[2] // 2]
    window_fft_im[
        c[0] - size[0] // 2 : c[0] + size[0] // 2,
        c[1] - size[1] // 2 : c[1] + size[1] // 2,
        c[2] - size[2] // 2 : c[2] + size[2] // 2,
    ] = (
        crop_center(fft_im, size)
        - window_fft_im[
            c[0] - size[0] // 2 : c[0] + size[0] // 2,
            c[1] - size[1] // 2 : c[1] + size[1] // 2,
            c[2] - size[2] // 2 : c[2] + size[2] // 2,
        ]
    )
    return np.fft.ifftn(np.fft.ifftshift(window_fft_im))


def projection_z_axis(im):
    """project image along z axis"""
    return np.sum(im, axis=0)


def pad_all_channels_of_image(im, pad_vals):
    nb_channels = im.shape[2]
    shape_padded = np.shape(np.pad(im[:, :, 0], pad_vals))
    im_padded = np.zeros((*shape_padded, 3))
    for c in range(nb_channels):
        im_c_padded = np.pad(im[:, :, c], pad_vals)
        im_padded[:, :, c] = im_c_padded
    print("shape", im_padded.shape)
    return im_padded


def stopping_criteria(energies, eps, n=4, m=2):
    if len(energies) <= n:
        return False
    mean_current = np.mean(energies[-m:])
    mean_old = np.mean(energies[-n : -n + m])
    return mean_old - mean_current < eps



from typing import TYPE_CHECKING

import numpy as np

from spfluo.utils.array import array_namespace
from spfluo.utils.volume import affine_transform as affine_transform_spfluo

if TYPE_CHECKING:
    from spfluo.utils.array import Array

interp_order = 3


def rotation(
    volumes: "Array", transform: "Array", order: int = 1, inverse: bool = False
):
    """
    transform: homogeneous transform
    """
    xp = array_namespace(volumes, transform)
    if not inverse:  # scipy's affine_transform do inverse transform by default
        transform = xp.linalg.inv(transform)
    return affine_transform_spfluo(
        volumes, transform, order=order, prefilter=False, batch=True
    )


def conversion_2_first_eulers_angles_cartesian(theta, phi, degrees=True):
    if degrees:
        theta = theta * np.pi / 180
        phi = phi * np.pi / 180
    x, y, z = np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)
    return x, y, z



from __future__ import annotations

import os
from typing import Dict, Tuple

import numpy as np

from ..utils import summary
from .core import Data, F1Score, Logger, Losses, fit, make_model
from .prepare_fs_data import make_dataloaders as make_fs_dataloaders
from .prepare_pu_data import make_dataloaders as make_pu_dataloaders


def make_data(
    # base params
    mode: str,
    rootdir: str,
    output_dir: str,
    batch_size: int,
    num_workers: int,
    num_pos_samples: int,
    augment: bool,
    shuffle: bool,
    extension: str,
    dim: int,
    epoch_size: int = None,
    # pu optional params
    radius: int = None,
    num_particles_per_image: int = None,
    patch_size: Tuple[int] = None,
    load_u_masks: bool = False,
) -> Data:
    args = [
        rootdir,
        output_dir,
        batch_size,
        num_workers,
        num_pos_samples,
        shuffle,
        augment,
        epoch_size,
        dim,
    ]
    maker = make_fs_dataloaders
    if mode == "pu":
        args += [extension, radius, num_particles_per_image, patch_size, load_u_masks]
        maker = make_pu_dataloaders
    train_loader, val_loader = maker(*args)
    return Data(train_loader, val_loader)


def make_logger(config: Dict) -> Logger:
    # Names order MUST match MultiTaskLoss outputs order.
    loss_names = ["classification"]
    if config["ae"]:
        loss_names.append("reconstruction")
    loss_names.append("total")
    losses = Losses(loss_names)
    f1_score = F1Score()
    return Logger(losses, f1_score, config)


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def train(
    # base params
    mode: str,
    rootdir: str,
    batch_size: int,
    num_workers: int,
    output_dir: str,
    num_epochs: int,
    lr: float,
    network: str,
    width_factor: float,
    depth_factor: float,
    swa: bool,
    swa_lr: float,
    patch_size: Tuple[int],
    num_pos_samples: int,
    shuffle: bool,
    augment: bool,
    dim: int,
    extension: str,
    epoch_size: int = None,
    downscale: float = 1.0,
    # pu optional params
    radius: int | None = None,
    num_particles_per_image: int = None,
    load_u_masks: bool = False,
    # autoencoder params
    ae: bool = False,
    eta: float = 0.0,
    learn_eta: bool = False,
    num_features: Tuple[int] = None,
    hidden_dim: int = None,
) -> None:
    print("\nPreparing training pipeline ...\n")
    config = locals()
    os.makedirs(output_dir, exist_ok=True)
    summary_output = os.path.join(output_dir, "train_config.txt")
    summary(config, title="TRAINING CONFIG", output=summary_output)
    patch_size = np.rint(np.array(patch_size) / downscale).astype(int)
    radius = round(radius / downscale)
    data = make_data(
        mode,
        rootdir,
        output_dir,
        batch_size,
        num_workers,
        num_pos_samples,
        augment,
        shuffle,
        extension,
        dim,
        epoch_size,
        radius,
        num_particles_per_image,
        patch_size,
        load_u_masks,
    )
    num_classes = 1
    pi = data.train_loader.sampler.pi if mode == "pu" else None
    train_params = [num_classes, dim, lr, patch_size, num_epochs, pi]
    model_params = [network, width_factor, depth_factor, swa, swa_lr]
    ae_params = [ae, eta, learn_eta, num_features, hidden_dim]
    model = make_model(*train_params, *model_params, *ae_params)
    logger = make_logger(config)
    fit(data, model, logger)



import os
import pickle
from typing import Dict, List, Tuple

import imageio
import numpy as np
from skimage import img_as_ubyte
from sklearn.metrics import classification_report, confusion_matrix

from ..utils import center_to_corners, load_annotations, load_array, summary
from .postprocess import compute_iou_3d, compute_volume

# +------------------------------------------------------------------------------------------+ #
# |                                       LABEL VOLUMES                                      | #
# +------------------------------------------------------------------------------------------+ #


def grayscale_to_rgb(image: np.ndarray) -> np.ndarray:
    return np.stack((image,) * 3, axis=-1)


def recenter_if_needed(shape: Tuple[int], corners: np.ndarray) -> Tuple[int]:
    D, H, W = shape
    x_min, y_min, z_min, x_max, y_max, z_max = corners.astype(np.int)
    d, h, w = z_max - z_min, y_max - y_min, x_max - x_min
    z_min, y_min, x_min = max(0, z_min), max(0, y_min), max(0, x_min)
    z_max, y_max, x_max = z_min + d, y_min + h, x_min + w
    z_max, y_max, x_max = min(D, z_max), min(H, y_max), min(W, x_max)
    z_min, y_min, x_min = z_max - d, y_max - h, x_max - w
    return x_min, y_min, z_min, x_max, y_max, z_max


def draw_bbox(
    image: np.ndarray, bbox: np.ndarray, color: Tuple[int] = (255, 0, 0)
) -> None:
    """Modify image inplace.
    This function draws a 3D bbox to be seen from the z-axis, that is a 2D bbox
    [x_min, x_max, y_min, y_max] at each z in range [z_min, z_max].
    """
    x_min, y_min, z_min, x_max, y_max, z_max = recenter_if_needed(image.shape[:3], bbox)
    image[z_min:z_max, y_min, x_min:x_max] = color
    image[z_min:z_max, y_max, x_min:x_max] = color
    image[z_min:z_max, y_min:y_max, x_min] = color
    image[z_min:z_max, y_min:y_max, x_max] = color


def draw_bboxes(
    image: np.ndarray, bboxes: np.ndarray, color: Tuple[int] = (255, 0, 0)
) -> None:
    """Modify image inplace."""
    for bbox in bboxes:
        draw_bbox(image, bbox, color)


def save_3d_image_to_tiff(image: np.ndarray, name: str) -> None:
    imageio.mimwrite(f"{name}.tiff", image)


def create_and_save_labelled_volume(
    predicted_corners: np.ndarray,
    true_corners: np.ndarray,
    tp: List[int],
    fp: List[int],
    fn: List[int],
    image_dir: str,
    image_name: str,
    output_dir: str,
) -> None:
    # 1. Load and prepare image array
    image = load_array(os.path.join(image_dir, image_name))
    image = grayscale_to_rgb(img_as_ubyte(image))
    # 2. Draw colored groundtruth and predictions
    colors = {"red": (255, 0, 0), "green": (0, 255, 0), "yellow": (255, 255, 0)}
    draw_bboxes(image, true_corners[fn], color=colors["red"])
    draw_bboxes(image, predicted_corners[tp], color=colors["green"])
    draw_bboxes(image, predicted_corners[fp], color=colors["yellow"])
    # 3. Save results to npz and tiff
    output_prefix = os.path.join(output_dir, image_name)
    np.savez(output_prefix, image=image)
    save_3d_image_to_tiff(image, output_prefix)


# +------------------------------------------------------------------------------------------+ #
# |                                    METRICS COMPUTATION                                   | #
# +------------------------------------------------------------------------------------------+ #


def get_predictions_indices(
    predictions: np.ndarray,
    groundtruth: np.ndarray,
    threshold: float,
) -> Dict:
    true_positive, false_positive, false_negative = [], [], []
    predictions_volume = compute_volume(predictions)
    groundtruth_volume = compute_volume(groundtruth)
    for i, prediction in enumerate(predictions):
        ious = compute_iou_3d(
            prediction, groundtruth, predictions_volume[i], groundtruth_volume
        )
        if ious.max() < threshold:
            false_positive.append(i)
    for i, gt in enumerate(groundtruth):
        ious = compute_iou_3d(
            gt, predictions, groundtruth_volume[i], predictions_volume
        )
        if ious.max() > threshold:
            true_positive.append(ious.argmax())
        else:
            false_negative.append(i)
    return true_positive, false_positive, false_negative


def evaluate_one_image_picking(
    predicted_corners: np.ndarray,
    true_corners: np.ndarray,
    threshold: float,
    label_volume: bool,
    image_dir: str = None,
    image_name: str = None,
    output_dir: str = None,
) -> Dict:
    tp, fp, fn = get_predictions_indices(predicted_corners, true_corners, threshold)
    if label_volume:
        label_args = [predicted_corners, true_corners, tp, fp, fn]
        io_args = [image_dir, image_name, output_dir]
        create_and_save_labelled_volume(*label_args, *io_args)
    true_positive, false_positive, false_negative = len(tp), len(fp), len(fn)
    precision = true_positive / (true_positive + false_positive + 1e-20)
    recall = true_positive / (true_positive + false_negative + 1e-20)
    f1 = 2 * precision * recall / (precision + recall + 1e-20)
    metrics = {
        "true_positive": true_positive,
        "false_positive": false_positive,
        "false_negative": false_negative,
        "precision": precision,
        "recall": recall,
        "f1": f1,
    }
    return metrics


# +------------------------------------------------------------------------------------------+ #
# |                                      EXTERNAL CALLS                                      | #
# +------------------------------------------------------------------------------------------+ #


def evaluate_picking(
    predictions_path: str,
    annotations_path: str,
    patch_size: str,
    threshold: float,
    step: str,
    make_labelled_volumes: bool,
    image_dir: str = None,
    image_name: str = None,
    output_dir: str = None,
) -> Dict[str, Dict[str, float]]:
    """If image_name is None, evaluate all predicted images.
    predictions[image_name] = {
        'raw':                       (N1', (x1, y1, z1, x2, y2, z2)),
        'nms':                       (N2', (x1, y1, z1, x2, y2, z2)),
        'first_spatial_average':     (N3', (x1, y1, z1, x2, y2, z2)),
        'center_of_mass_correction': (N4', (x1, y1, z1, x2, y2, z2)),
        'last_spatial_average':      (N5', (x1, y1, z1, x2, y2, z2)),
    }
    with N1 > N2 > N3 = N4 > N5.
    See postprocess.py.
    """
    summary_output = None
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)
        summary_output = os.path.join(output_dir, "evaluation_config.txt")
    summary(locals(), title="EVALUATION", output=summary_output)
    with open(predictions_path, "rb") as file:
        predictions = pickle.load(file)
    annotations = load_annotations(annotations_path)
    images_names = [image_name] if image_name is not None else predictions.keys()
    evaluations = {}
    volumes_output_dir = None
    if make_labelled_volumes:
        volumes_output_dir = os.path.join(output_dir, "labelled_volumes")
        os.makedirs(volumes_output_dir, exist_ok=True)
    for image_name in images_names:
        true_centers = annotations[annotations[:, 0] == image_name, 1:]
        true_corners = [
            center_to_corners(center, patch_size) for center in true_centers
        ]
        true_corners = np.array(true_corners)
        predicted_corners = predictions[image_name][step]
        metrics_args = [predicted_corners, true_corners, threshold]
        label_args = [make_labelled_volumes, image_dir, image_name, volumes_output_dir]
        evaluations[image_name] = evaluate_one_image_picking(*metrics_args, *label_args)
    mean_metrics_over_testset = {}
    for metric in [
        "true_positive",
        "false_positive",
        "false_negative",
        "precision",
        "recall",
        "f1",
    ]:
        values = [evaluations[image_name][metric] for image_name in images_names]
        mean_metrics_over_testset[metric] = np.array(values).mean()
    if output_dir is not None:
        summary_output = os.path.join(output_dir, "evaluation_results.txt")
        summary(
            mean_metrics_over_testset,
            title="EVALUATION MEAN RESULTS",
            output=summary_output,
        )
        with open(os.path.join(output_dir, "evaluations.pickle"), "wb") as file:
            pickle.dump(evaluations, file)
    return evaluations


def flatten_dict(dd, separator=".", prefix=""):
    # https://stackoverflow.com/a/19647596
    return (
        {
            prefix + separator + k if prefix else k: v
            for kk, vv in dd.items()
            for k, v in flatten_dict(vv, separator, kk).items()
        }
        if isinstance(dd, dict)
        else {prefix: dd}
    )


def evaluate_tilt(
    predictions_path: str,
    annotations_path: str,
    image_name: str = None,
    output_dir: str = None,
    selected_particules: str = None,  # particules from the training set
) -> Dict[str, Dict[str, float]]:
    if selected_particules is not None:
        with open(selected_particules, "rb") as f:
            train_particles = pickle.load(f)
    else:
        train_particles = None
    summary_output = None
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)
        summary_output = os.path.join(output_dir, "evaluation_config.txt")
    # summary(locals(), title='EVALUATION', output=summary_output)
    with open(predictions_path, "rb") as file:
        predictions = pickle.load(file)
    annotations = load_annotations(annotations_path)
    images_names = [image_name] if image_name is not None else predictions.keys()
    evaluations = {}
    for image_name in images_names:
        if train_particles is not None:
            train_particles_image = train_particles[
                train_particles[:, 1] == image_name, 0
            ]
        else:
            train_particles_image = []
        true_views = (
            annotations[annotations[:, 0] == image_name, 1:].reshape(-1).astype(int)
        )
        predicted_views = predictions[image_name][0].astype(int)
        if predicted_views.shape[0] > true_views.shape[0]:
            predicted_views = predicted_views[: true_views.shape[0]]
        evaluations[image_name] = {}
        evaluations[image_name]["train"] = np.array(
            [i in train_particles_image for i in range(len(true_views))], dtype=bool
        )  # train mask
        evaluations[image_name]["true"] = true_views
        evaluations[image_name]["pred"] = predicted_views

    eval = {
        k: np.concatenate([evaluations[im][k] for im in images_names])
        for k in ["train", "true", "pred"]
    }

    y_test_true = eval["true"][~eval["train"]]
    y_train_true = eval["true"][eval["train"]]
    y_test_pred = eval["pred"][~eval["train"]]
    y_train_pred = eval["pred"][eval["train"]]

    if selected_particules is not None:
        cr_train = classification_report(
            y_train_true,
            y_train_pred,
            labels=[0, 1, 2],
            target_names=["top", "side", "other"],
        )
        cm_train = confusion_matrix(y_train_true, y_train_pred, labels=[0, 1, 2])
    cr_test = classification_report(
        y_test_true,
        y_test_pred,
        labels=[0, 1, 2],
        target_names=["top", "side", "other"],
    )
    cm_test = confusion_matrix(y_test_true, y_test_pred, labels=[0, 1, 2])

    if output_dir is not None:
        summary_output = os.path.join(output_dir, "evaluation_results.txt")
        with open(summary_output, "w") as f:
            if selected_particules is not None:
                f.write("TRAIN SET\n\n")
                f.write(str(cm_train) + "\n\n")
                f.write(cr_train + "\n\n")
            f.write("TEST SET\n\n")
            f.write(str(cm_test) + "\n\n")
            f.write(cr_test + "\n\n")

        with open(os.path.join(output_dir, "evaluations.pickle"), "wb") as file:
            pickle.dump(evaluations, file)
    return evaluations



import hashlib
import sys
import zipfile
from contextlib import contextmanager
from pathlib import Path

import requests

from spfluo.data.constants import (
    ARCHIVE_NAME,
    REPO_ID,
    SEAFILE_URL,
    UNISTRA_PASSWORD,
    UNISTRA_USERNAME,
)

if sys.version_info >= (3, 9):
    from importlib.resources import as_file, files

    @contextmanager
    def _get_registry_file():
        with as_file(files("spfluo").joinpath("data/registry.txt")) as registry:
            yield registry
else:

    @contextmanager
    def _get_registry_file():
        registry = Path(__file__).parent.joinpath("data/registry.txt")
        yield registry


def _file_hash(file: Path):
    return hashlib.sha256(file.read_bytes()).hexdigest()


def get_token():
    # Step 1: Authenticate and get token
    auth_response = requests.post(
        f"{SEAFILE_URL}/api2/auth-token/",
        data={
            "username": f"{UNISTRA_USERNAME}@unistra.fr",
            "password": UNISTRA_PASSWORD,
        },
    )
    auth_response.raise_for_status()  # This will raise an error if the request failed
    return auth_response.json().get("token")


def upload_archive():
    # Step 0: Create a zip archive from registry.txt
    with zipfile.ZipFile(ARCHIVE_NAME, "w") as zipf:
        with _get_registry_file() as registry_path:
            with open(registry_path, "r") as registry:
                for line in registry:
                    # Assuming space separates the path and the hash
                    file, hash = line.strip().split(" ")
                    file_path = registry_path.parent / file
                    print(f"Adding {file_path} to zip")
                    if file_path.exists() and _file_hash(file_path) == hash:
                        zipf.write(file_path, file_path.relative_to(registry_path.parent))
                    else:
                        raise FileNotFoundError(
                            f"{file_path} was not found or hash was wrong"
                        )

    # Step 2: Get upload link
    headers = {"Authorization": f"Token {get_token()}"}
    upload_link_response = requests.get(
        f"{SEAFILE_URL}/api2/repos/{REPO_ID}/upload-link/", headers=headers
    )
    upload_link_response.raise_for_status()
    # Removing quotes from the response
    upload_link = upload_link_response.text.strip('"')

    # Step 3: Upload file
    with open(ARCHIVE_NAME, "rb") as file:
        f = {"file": file, "parent_dir": ("", "/"), "replace": (None, "1")}
        print("Uploading file to seafile...")
        upload_response = requests.post(upload_link, headers=headers, files=f)
        upload_response.raise_for_status()

    print("File uploaded successfully.")


if __name__ == "__main__":
    upload_archive()
